# Copyright (C) 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""A simple first-in-first-out (FIFO) cache."""

from collections import deque
from typing import Any, Callable, Deque, Dict


class FIFOCache(dict):
    """A class which manages a cache of entries, removing old ones."""

    def __init__(self, max_cache: int = 100, after_cleanup_count=None) -> None:
        dict.__init__(self)
        self._max_cache = max_cache
        if after_cleanup_count is None:
            self._after_cleanup_count = self._max_cache * 8 // 10
        else:
            self._after_cleanup_count = min(after_cleanup_count,
                                            self._max_cache)
        self._cleanup: Dict[Any, Callable[[], None]] = {}  # map to cleanup functions when items are removed
        self._queue: Deque[Any] = deque()  # Track when things are accessed

    def __setitem__(self, key, value):
        """Add a value to the cache, there will be no cleanup function."""
        self.add(key, value, cleanup=None)

    def __delitem__(self, key):
        # Remove the key from an arbitrary location in the queue
        self._queue.remove(key)
        self._remove(key)

    def add(self, key, value, cleanup=None):
        """Add a new value to the cache.

        Also, if the entry is ever removed from the queue, call cleanup.
        Passing it the key and value being removed.

        :param key: The key to store it under
        :param value: The object to store
        :param cleanup: None or a function taking (key, value) to indicate
                        'value' should be cleaned up
        """
        if key in self:
            # Remove the earlier reference to this key, adding it again bumps
            # it to the end of the queue
            del self[key]
        self._queue.append(key)
        dict.__setitem__(self, key, value)
        if cleanup is not None:
            self._cleanup[key] = cleanup
        if len(self) > self._max_cache:
            self.cleanup()

    def cache_size(self):
        """Get the number of entries we will cache."""
        return self._max_cache

    def cleanup(self):
        """Clear the cache until it shrinks to the requested size.

        This does not completely wipe the cache, just makes sure it is under
        the after_cleanup_count.
        """
        # Make sure the cache is shrunk to the correct size
        while len(self) > self._after_cleanup_count:
            self._remove_oldest()
        if len(self._queue) != len(self):
            raise AssertionError('The length of the queue should always equal'
                                 ' the length of the dict. %s != %s'
                                 % (len(self._queue), len(self)))

    def clear(self):
        """Clear out all of the cache."""
        # Clean up in FIFO order
        while self:
            self._remove_oldest()

    def _remove(self, key):
        """Remove an entry, making sure to call any cleanup function."""
        cleanup = self._cleanup.pop(key, None)
        # We override self.pop() because it doesn't play well with cleanup
        # functions.
        val = dict.pop(self, key)
        if cleanup is not None:
            cleanup(key, val)
        return val

    def _remove_oldest(self):
        """Remove the oldest entry."""
        key = self._queue.popleft()
        self._remove(key)

    def resize(self, max_cache, after_cleanup_count=None):
        """Increase/decrease the number of cached entries.

        :param max_cache: The maximum number of entries to cache.
        :param after_cleanup_count: After cleanup, we should have at most this
            many entries. This defaults to 80% of max_cache.
        """
        self._max_cache = max_cache
        if after_cleanup_count is None:
            self._after_cleanup_count = max_cache * 8 // 10
        else:
            self._after_cleanup_count = min(max_cache, after_cleanup_count)
        if len(self) > self._max_cache:
            self.cleanup()

    # raise NotImplementedError on dict functions that would mutate the cache
    # which have not been properly implemented yet.
    def copy(self):
        raise NotImplementedError(self.copy)

    def pop(self, key, default=None):
        # If there is a cleanup() function, than it is unclear what pop()
        # should do. Specifically, we would have to call the cleanup on the
        # value before we return it, which should cause whatever resources were
        # allocated to be removed, which makes the return value fairly useless.
        # So instead, we just don't implement it.
        raise NotImplementedError(self.pop)

    def popitem(self):
        # See pop()
        raise NotImplementedError(self.popitem)

    def setdefault(self, key, defaultval=None):
        """similar to dict.setdefault"""
        if key in self:
            return self[key]
        self[key] = defaultval
        return defaultval

    def update(self, *args, **kwargs):
        """Similar to dict.update()"""
        if len(args) == 1:
            arg = args[0]
            if isinstance(arg, dict):
                for key in arg:
                    self.add(key, arg[key])
            else:
                for key, val in args[0]:
                    self.add(key, val)
        elif len(args) > 1:
            raise TypeError('update expected at most 1 argument, got %d'
                            % len(args))
        if kwargs:
            for key in kwargs:
                self.add(key, kwargs[key])


class FIFOSizeCache(FIFOCache):
    """An FIFOCache that removes things based on the size of the values.

    This differs in that it doesn't care how many actual items there are,
    it restricts the cache to be cleaned based on the size of the data.
    """

    def __init__(self, max_size=1024 * 1024, after_cleanup_size=None,
                 compute_size=None):
        """Create a new FIFOSizeCache.

        :param max_size: The max number of bytes to store before we start
            clearing out entries.
        :param after_cleanup_size: After cleaning up, shrink everything to this
            size (defaults to 80% of max_size).
        :param compute_size: A function to compute the size of a value. If
            not supplied we default to 'len'.
        """
        # Arbitrary, we won't really be using the value anyway.
        FIFOCache.__init__(self, max_cache=max_size)
        self._max_size = max_size
        if after_cleanup_size is None:
            self._after_cleanup_size = self._max_size * 8 // 10
        else:
            self._after_cleanup_size = min(after_cleanup_size, self._max_size)

        self._value_size = 0
        self._compute_size = compute_size
        if compute_size is None:
            self._compute_size = len

    def add(self, key, value, cleanup=None):
        """Add a new value to the cache.

        Also, if the entry is ever removed from the queue, call cleanup.
        Passing it the key and value being removed.

        :param key: The key to store it under
        :param value: The object to store, this value by itself is >=
            after_cleanup_size, then we will not store it at all.
        :param cleanup: None or a function taking (key, value) to indicate
                        'value' sohuld be cleaned up.
        """
        # Even if the new value won't be stored, we need to remove the old
        # value
        if key in self:
            # Remove the earlier reference to this key, adding it again bumps
            # it to the end of the queue
            del self[key]
        value_len = self._compute_size(value)
        if value_len >= self._after_cleanup_size:
            return
        self._queue.append(key)
        dict.__setitem__(self, key, value)
        if cleanup is not None:
            self._cleanup[key] = cleanup
        self._value_size += value_len
        if self._value_size > self._max_size:
            # Time to cleanup
            self.cleanup()

    def cache_size(self):
        """Get the number of bytes we will cache."""
        return self._max_size

    def cleanup(self):
        """Clear the cache until it shrinks to the requested size.

        This does not completely wipe the cache, just makes sure it is under
        the after_cleanup_size.
        """
        # Make sure the cache is shrunk to the correct size
        while self._value_size > self._after_cleanup_size:
            self._remove_oldest()

    def _remove(self, key):
        """Remove an entry, making sure to maintain the invariants."""
        val = FIFOCache._remove(self, key)
        self._value_size -= self._compute_size(val)
        return val

    def resize(self, max_size, after_cleanup_size=None):
        """Increase/decrease the amount of cached data.

        :param max_size: The maximum number of bytes to cache.
        :param after_cleanup_size: After cleanup, we should have at most this
            many bytes cached. This defaults to 80% of max_size.
        """
        FIFOCache.resize(self, max_size)
        self._max_size = max_size
        if after_cleanup_size is None:
            self._after_cleanup_size = max_size * 8 // 10
        else:
            self._after_cleanup_size = min(max_size, after_cleanup_size)
        if self._value_size > self._max_size:
            self.cleanup()



# Copyright (C) 2005-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Test testaments for gpg signing."""

# TODO: Testaments with x-bits

import os

from breezy import osutils
from breezy.bzr.testament import StrictTestament, StrictTestament3, Testament
from breezy.tests import TestCaseWithTransport
from breezy.tests.features import SymlinkFeature


class TestamentSetup(TestCaseWithTransport):

    def setUp(self):
        super().setUp()
        self.wt = self.make_branch_and_tree('.')
        self.wt.set_root_id(b'TREE_ROT')
        b = self.b = self.wt.branch
        b.nick = "test branch"
        self.wt.commit(message='initial null commit',
                       committer='test@user',
                       timestamp=1129025423,  # 'Tue Oct 11 20:10:23 2005'
                       timezone=0,
                       rev_id=b'test@user-1')
        self.build_tree_contents([('hello', b'contents of hello file'),
                                  ('src/', ),
                                  ('src/foo.c', b'int main()\n{\n}\n')])
        self.wt.add(['hello', 'src', 'src/foo.c'],
                    ids=[b'hello-id', b'src-id', b'foo.c-id'])
        tt = self.wt.transform()
        trans_id = tt.trans_id_tree_path('hello')
        tt.set_executability(True, trans_id)
        tt.apply()
        self.wt.commit(message='add files and directories',
                       timestamp=1129025483,
                       timezone=36000,
                       rev_id=b'test@user-2',
                       committer='test@user')


class TestamentTests(TestamentSetup):

    def testament_class(self):
        return Testament

    def expected(self, key):
        return texts[self.testament_class()][key]

    def from_revision(self, repository, revision_id):
        return self.testament_class().from_revision(repository, revision_id)

    def test_null_testament(self):
        """Testament for a revision with no contents."""
        t = self.from_revision(self.b.repository, b'test@user-1')
        ass = self.assertTrue
        eq = self.assertEqual
        ass(isinstance(t, Testament))
        eq(t.revision_id, b'test@user-1')
        eq(t.committer, 'test@user')
        eq(t.timestamp, 1129025423)
        eq(t.timezone, 0)

    def test_testment_text_form(self):
        """Conversion of testament to canonical text form."""
        t = self.from_revision(self.b.repository, b'test@user-1')
        text_form = t.as_text()
        self.log('testament text form:\n%s' % text_form)
        self.assertEqualDiff(text_form, self.expected('rev_1'))
        short_text_form = t.as_short_text()
        self.assertEqualDiff(short_text_form, self.expected('rev_1_short'))

    def test_testament_with_contents(self):
        """Testament containing a file and a directory."""
        t = self.from_revision(self.b.repository, b'test@user-2')
        text_form = t.as_text()
        self.log('testament text form:\n%s' % text_form)
        self.assertEqualDiff(text_form, self.expected('rev_2'))
        actual_short = t.as_short_text()
        self.assertEqualDiff(actual_short, self.expected('rev_2_short'))

    def test_testament_symlinks(self):
        """Testament containing symlink (where possible)"""
        self.requireFeature(SymlinkFeature(self.test_dir))
        os.symlink('wibble/linktarget', 'link')
        self.wt.add(['link'], ids=[b'link-id'])
        self.wt.commit(message='add symlink',
                       timestamp=1129025493,
                       timezone=36000,
                       rev_id=b'test@user-3',
                       committer='test@user')
        t = self.from_revision(self.b.repository, b'test@user-3')
        self.assertEqualDiff(t.as_text(), self.expected('rev_3'))

    def test_testament_revprops(self):
        """Testament to revision with extra properties"""
        props = {'flavor': 'sour cherry\ncream cheese',
                 'size': 'medium',
                 'empty': '',
                 }
        self.wt.commit(message='revision with properties',
                       timestamp=1129025493,
                       timezone=36000,
                       rev_id=b'test@user-3',
                       committer='test@user',
                       revprops=props)
        t = self.from_revision(self.b.repository, b'test@user-3')
        self.assertEqualDiff(t.as_text(), self.expected('rev_props'))

    def test_testament_unicode_commit_message(self):
        self.wt.commit(
            message='non-ascii commit \N{COPYRIGHT SIGN} me',
            timestamp=1129025493,
            timezone=36000,
            rev_id=b'test@user-3',
            committer='Erik B\xe5gfors <test@user>',
            revprops={'uni': '\xb5'}
            )
        t = self.from_revision(self.b.repository, b'test@user-3')
        self.assertEqualDiff(
            self.expected('sample_unicode').encode('utf-8'), t.as_text())

    def test_from_tree(self):
        tree = self.b.repository.revision_tree(b'test@user-2')
        testament = self.testament_class().from_revision_tree(tree)
        text_1 = testament.as_short_text()
        text_2 = self.from_revision(self.b.repository,
                                    b'test@user-2').as_short_text()
        self.assertEqual(text_1, text_2)

    def test___init__(self):
        revision = self.b.repository.get_revision(b'test@user-2')
        tree = self.b.repository.revision_tree(b'test@user-2')
        testament_1 = self.testament_class()(revision, tree)
        text_1 = testament_1.as_short_text()
        text_2 = self.from_revision(self.b.repository,
                                    b'test@user-2').as_short_text()
        self.assertEqual(text_1, text_2)


class TestamentTestsStrict(TestamentTests):

    def testament_class(self):
        return StrictTestament


class TestamentTestsStrict2(TestamentTests):

    def testament_class(self):
        return StrictTestament3


REV_1_TESTAMENT = b"""\
bazaar-ng testament version 1
revision-id: test@user-1
committer: test@user
timestamp: 1129025423
timezone: 0
parents:
message:
  initial null commit
inventory:
properties:
  branch-nick:
    test branch
"""


REV_1_STRICT_TESTAMENT = b"""\
bazaar-ng testament version 2.1
revision-id: test@user-1
committer: test@user
timestamp: 1129025423
timezone: 0
parents:
message:
  initial null commit
inventory:
properties:
  branch-nick:
    test branch
"""


REV_1_STRICT_TESTAMENT3 = b"""\
bazaar testament version 3 strict
revision-id: test@user-1
committer: test@user
timestamp: 1129025423
timezone: 0
parents:
message:
  initial null commit
inventory:
  directory . TREE_ROT test@user-1 no
properties:
  branch-nick:
    test branch
"""


REV_1_SHORT = b"""\
bazaar-ng testament short form 1
revision-id: test@user-1
sha1: %s
""" % osutils.sha_string(REV_1_TESTAMENT)


REV_1_SHORT_STRICT = b"""\
bazaar-ng testament short form 2.1
revision-id: test@user-1
sha1: %s
""" % osutils.sha_string(REV_1_STRICT_TESTAMENT)


REV_1_SHORT_STRICT3 = b"""\
bazaar testament short form 3 strict
revision-id: test@user-1
sha1: %s
""" % osutils.sha_string(REV_1_STRICT_TESTAMENT3)


REV_2_TESTAMENT = b"""\
bazaar-ng testament version 1
revision-id: test@user-2
committer: test@user
timestamp: 1129025483
timezone: 36000
parents:
  test@user-1
message:
  add files and directories
inventory:
  file hello hello-id 34dd0ac19a24bf80c4d33b5c8960196e8d8d1f73
  directory src src-id
  file src/foo.c foo.c-id a2a049c20f908ae31b231d98779eb63c66448f24
properties:
  branch-nick:
    test branch
"""


REV_2_STRICT_TESTAMENT = b"""\
bazaar-ng testament version 2.1
revision-id: test@user-2
committer: test@user
timestamp: 1129025483
timezone: 36000
parents:
  test@user-1
message:
  add files and directories
inventory:
  file hello hello-id 34dd0ac19a24bf80c4d33b5c8960196e8d8d1f73 test@user-2 yes
  directory src src-id test@user-2 no
  file src/foo.c foo.c-id a2a049c20f908ae31b231d98779eb63c66448f24 test@user-2 no
properties:
  branch-nick:
    test branch
"""


REV_2_STRICT_TESTAMENT3 = b"""\
bazaar testament version 3 strict
revision-id: test@user-2
committer: test@user
timestamp: 1129025483
timezone: 36000
parents:
  test@user-1
message:
  add files and directories
inventory:
  directory . TREE_ROT test@user-1 no
  file hello hello-id 34dd0ac19a24bf80c4d33b5c8960196e8d8d1f73 test@user-2 yes
  directory src src-id test@user-2 no
  file src/foo.c foo.c-id a2a049c20f908ae31b231d98779eb63c66448f24 test@user-2 no
properties:
  branch-nick:
    test branch
"""


REV_2_SHORT = b"""\
bazaar-ng testament short form 1
revision-id: test@user-2
sha1: %s
""" % osutils.sha_string(REV_2_TESTAMENT)


REV_2_SHORT_STRICT = b"""\
bazaar-ng testament short form 2.1
revision-id: test@user-2
sha1: %s
""" % osutils.sha_string(REV_2_STRICT_TESTAMENT)


REV_2_SHORT_STRICT3 = b"""\
bazaar testament short form 3 strict
revision-id: test@user-2
sha1: %s
""" % osutils.sha_string(REV_2_STRICT_TESTAMENT3)


REV_PROPS_TESTAMENT = b"""\
bazaar-ng testament version 1
revision-id: test@user-3
committer: test@user
timestamp: 1129025493
timezone: 36000
parents:
  test@user-2
message:
  revision with properties
inventory:
  file hello hello-id 34dd0ac19a24bf80c4d33b5c8960196e8d8d1f73
  directory src src-id
  file src/foo.c foo.c-id a2a049c20f908ae31b231d98779eb63c66448f24
properties:
  branch-nick:
    test branch
  empty:
  flavor:
    sour cherry
    cream cheese
  size:
    medium
"""


REV_PROPS_TESTAMENT_STRICT = b"""\
bazaar-ng testament version 2.1
revision-id: test@user-3
committer: test@user
timestamp: 1129025493
timezone: 36000
parents:
  test@user-2
message:
  revision with properties
inventory:
  file hello hello-id 34dd0ac19a24bf80c4d33b5c8960196e8d8d1f73 test@user-2 yes
  directory src src-id test@user-2 no
  file src/foo.c foo.c-id a2a049c20f908ae31b231d98779eb63c66448f24 test@user-2 no
properties:
  branch-nick:
    test branch
  empty:
  flavor:
    sour cherry
    cream cheese
  size:
    medium
"""


REV_PROPS_TESTAMENT_STRICT3 = b"""\
bazaar testament version 3 strict
revision-id: test@user-3
committer: test@user
timestamp: 1129025493
timezone: 36000
parents:
  test@user-2
message:
  revision with properties
inventory:
  directory . TREE_ROT test@user-1 no
  file hello hello-id 34dd0ac19a24bf80c4d33b5c8960196e8d8d1f73 test@user-2 yes
  directory src src-id test@user-2 no
  file src/foo.c foo.c-id a2a049c20f908ae31b231d98779eb63c66448f24 test@user-2 no
properties:
  branch-nick:
    test branch
  empty:
  flavor:
    sour cherry
    cream cheese
  size:
    medium
"""


REV_3_TESTAMENT = b"""\
bazaar-ng testament version 1
revision-id: test@user-3
committer: test@user
timestamp: 1129025493
timezone: 36000
parents:
  test@user-2
message:
  add symlink
inventory:
  file hello hello-id 34dd0ac19a24bf80c4d33b5c8960196e8d8d1f73
  symlink link link-id wibble/linktarget
  directory src src-id
  file src/foo.c foo.c-id a2a049c20f908ae31b231d98779eb63c66448f24
properties:
  branch-nick:
    test branch
"""


REV_3_TESTAMENT_STRICT = b"""\
bazaar-ng testament version 2.1
revision-id: test@user-3
committer: test@user
timestamp: 1129025493
timezone: 36000
parents:
  test@user-2
message:
  add symlink
inventory:
  file hello hello-id 34dd0ac19a24bf80c4d33b5c8960196e8d8d1f73 test@user-2 yes
  symlink link link-id wibble/linktarget test@user-3 no
  directory src src-id test@user-2 no
  file src/foo.c foo.c-id a2a049c20f908ae31b231d98779eb63c66448f24 test@user-2 no
properties:
  branch-nick:
    test branch
"""


REV_3_TESTAMENT_STRICT3 = b"""\
bazaar testament version 3 strict
revision-id: test@user-3
committer: test@user
timestamp: 1129025493
timezone: 36000
parents:
  test@user-2
message:
  add symlink
inventory:
  directory . TREE_ROT test@user-1 no
  file hello hello-id 34dd0ac19a24bf80c4d33b5c8960196e8d8d1f73 test@user-2 yes
  symlink link link-id wibble/linktarget test@user-3 no
  directory src src-id test@user-2 no
  file src/foo.c foo.c-id a2a049c20f908ae31b231d98779eb63c66448f24 test@user-2 no
properties:
  branch-nick:
    test branch
"""


SAMPLE_UNICODE_TESTAMENT = """\
bazaar-ng testament version 1
revision-id: test@user-3
committer: Erik B\xe5gfors <test@user>
timestamp: 1129025493
timezone: 36000
parents:
  test@user-2
message:
  non-ascii commit \N{COPYRIGHT SIGN} me
inventory:
  file hello hello-id 34dd0ac19a24bf80c4d33b5c8960196e8d8d1f73
  directory src src-id
  file src/foo.c foo.c-id a2a049c20f908ae31b231d98779eb63c66448f24
properties:
  branch-nick:
    test branch
  uni:
    \xb5
"""


SAMPLE_UNICODE_TESTAMENT_STRICT = """\
bazaar-ng testament version 2.1
revision-id: test@user-3
committer: Erik B\xe5gfors <test@user>
timestamp: 1129025493
timezone: 36000
parents:
  test@user-2
message:
  non-ascii commit \N{COPYRIGHT SIGN} me
inventory:
  file hello hello-id 34dd0ac19a24bf80c4d33b5c8960196e8d8d1f73 test@user-2 yes
  directory src src-id test@user-2 no
  file src/foo.c foo.c-id a2a049c20f908ae31b231d98779eb63c66448f24 test@user-2 no
properties:
  branch-nick:
    test branch
  uni:
    \xb5
"""


SAMPLE_UNICODE_TESTAMENT_STRICT3 = """\
bazaar testament version 3 strict
revision-id: test@user-3
committer: Erik B\xe5gfors <test@user>
timestamp: 1129025493
timezone: 36000
parents:
  test@user-2
message:
  non-ascii commit \N{COPYRIGHT SIGN} me
inventory:
  directory . TREE_ROT test@user-1 no
  file hello hello-id 34dd0ac19a24bf80c4d33b5c8960196e8d8d1f73 test@user-2 yes
  directory src src-id test@user-2 no
  file src/foo.c foo.c-id a2a049c20f908ae31b231d98779eb63c66448f24 test@user-2 no
properties:
  branch-nick:
    test branch
  uni:
    \xb5
"""


texts = {
    Testament: {'rev_1': REV_1_TESTAMENT,
                'rev_1_short': REV_1_SHORT,
                'rev_2': REV_2_TESTAMENT,
                'rev_2_short': REV_2_SHORT,
                'rev_3': REV_3_TESTAMENT,
                'rev_props': REV_PROPS_TESTAMENT,
                'sample_unicode': SAMPLE_UNICODE_TESTAMENT,
                },
    StrictTestament: {'rev_1': REV_1_STRICT_TESTAMENT,
                      'rev_1_short': REV_1_SHORT_STRICT,
                      'rev_2': REV_2_STRICT_TESTAMENT,
                      'rev_2_short': REV_2_SHORT_STRICT,
                      'rev_3': REV_3_TESTAMENT_STRICT,
                      'rev_props': REV_PROPS_TESTAMENT_STRICT,
                      'sample_unicode': SAMPLE_UNICODE_TESTAMENT_STRICT,
                      },
    StrictTestament3: {'rev_1': REV_1_STRICT_TESTAMENT3,
                       'rev_1_short': REV_1_SHORT_STRICT3,
                       'rev_2': REV_2_STRICT_TESTAMENT3,
                       'rev_2_short': REV_2_SHORT_STRICT3,
                       'rev_3': REV_3_TESTAMENT_STRICT3,
                       'rev_props': REV_PROPS_TESTAMENT_STRICT3,
                       'sample_unicode': SAMPLE_UNICODE_TESTAMENT_STRICT3,
                       },
}



# Copyright (C) 2005, 2006, 2007, 2009, 2011, 2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Black-box tests for brz revno.
"""

import os

from breezy import tests
from breezy.bzr.tests.matchers import ContainsNoVfsCalls
from breezy.errors import NoSuchRevision


class TestRevno(tests.TestCaseWithTransport):

    def test_revno(self):

        def bzr(*args, **kwargs):
            return self.run_bzr(*args, **kwargs)[0]

        os.mkdir('a')
        os.chdir('a')
        bzr('init')
        self.assertEqual(int(bzr('revno')), 0)

        with open('foo', 'wb') as f:
            f.write(b'foo\n')
        bzr('add foo')
        bzr('commit -m foo')
        self.assertEqual(int(bzr('revno')), 1)

        os.mkdir('baz')
        bzr('add baz')
        bzr('commit -m baz')
        self.assertEqual(int(bzr('revno')), 2)

        os.chdir('..')
        self.assertEqual(int(bzr('revno a')), 2)
        self.assertEqual(int(bzr('revno a/baz')), 2)

    def test_revno_tree(self):
        # Make branch and checkout
        wt = self.make_branch_and_tree('branch')
        checkout = wt.branch.create_checkout('checkout', lightweight=True)

        # Get the checkout out of date
        self.build_tree(['branch/file'])
        wt.add(['file'])
        wt.commit('mkfile')

        # Make sure revno says we're on 1
        out, err = self.run_bzr('revno checkout')
        self.assertEqual('', err)
        self.assertEqual('1\n', out)

        # Make sure --tree knows it's still on 0
        out, err = self.run_bzr('revno --tree checkout')
        self.assertEqual('', err)
        self.assertEqual('0\n', out)

    def test_revno_tree_no_tree(self):
        # Make treeless branch
        b = self.make_branch('branch')

        # Try getting it's --tree revno
        out, err = self.run_bzr('revno --tree branch', retcode=3)
        self.assertEqual('', out)
        self.assertEqual('brz: ERROR: No WorkingTree exists for "branch".\n',
                         err)

    def test_dotted_revno_tree(self):
        builder = self.make_branch_builder('branch')
        builder.start_series()
        builder.build_snapshot(None, [
            ('add', ('', b'root-id', 'directory', None)),
            ('add', ('file', b'file-id', 'file', b'content\n'))],
            revision_id=b'A-id')
        builder.build_snapshot([b'A-id'], [], revision_id=b'B-id')
        builder.build_snapshot([b'A-id', b'B-id'], [], revision_id=b'C-id')
        builder.finish_series()
        b = builder.get_branch()
        co_b = b.create_checkout('checkout_b', lightweight=True,
                                 revision_id=b'B-id')
        out, err = self.run_bzr('revno checkout_b')
        self.assertEqual('', err)
        self.assertEqual('2\n', out)
        out, err = self.run_bzr('revno --tree checkout_b')
        self.assertEqual('', err)
        self.assertEqual('1.1.1\n', out)

    def test_stale_revno_tree(self):
        builder = self.make_branch_builder('branch')
        builder.start_series()
        builder.build_snapshot(None, [
            ('add', ('', b'root-id', 'directory', None)),
            ('add', ('file', b'file-id', 'file', b'content\n'))],
            revision_id=b'A-id')
        builder.build_snapshot([b'A-id'], [], revision_id=b'B-id')
        builder.build_snapshot([b'A-id'], [], revision_id=b'C-id')
        builder.finish_series()
        b = builder.get_branch()
        # The branch is now at "C-id", but the checkout is still at "B-id"
        # which is no longer in the history
        co_b = b.create_checkout('checkout_b', lightweight=True,
                                 revision_id=b'B-id')
        out, err = self.run_bzr('revno checkout_b')
        self.assertEqual('', err)
        self.assertEqual('2\n', out)
        out, err = self.run_bzr('revno --tree checkout_b')
        self.assertEqual('', err)
        self.assertEqual('???\n', out)

    def test_revno_ghost(self):
        builder = self.make_branch_builder('branch')
        builder.start_series()
        revid = builder.build_snapshot([b'aghost'], [
            ('add', ('', b'root-id', 'directory', None)),
            ('add', ('file', b'file-id', 'file', b'content\n'))],
            revision_id=b'A-id', allow_leftmost_as_ghost=True)
        builder.finish_series()
        b = builder.get_branch()

        def revision_id_to_revno(s, r):
            raise NoSuchRevision(s, r)
        self.overrideAttr(type(b), 'revision_id_to_dotted_revno', revision_id_to_revno)
        self.overrideAttr(type(b), 'revision_id_to_revno', revision_id_to_revno)
        out, err = self.run_bzr('revno branch')
        self.assertEqual('', err)
        self.assertEqual('???\n', out)

    def test_revno_with_revision(self):
        wt = self.make_branch_and_tree('.')
        revid1 = wt.commit('rev1')
        revid2 = wt.commit('rev2')

        out, err = self.run_bzr('revno -r-2 .')
        self.assertEqual('1\n', out)

        out, err = self.run_bzr('revno -rrevid:%s .' % revid1.decode('utf-8'))
        self.assertEqual('1\n', out)

    def test_revno_and_tree_mutually_exclusive(self):
        wt = self.make_branch_and_tree('.')
        out, err = self.run_bzr('revno -r-2 --tree .', retcode=3)
        self.assertEqual('', out)
        self.assertEqual(
            'brz: ERROR: --tree and --revision can not be used together\n',
            err)



# Copyright (C) 2006-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""WSGI application for bzr HTTP smart server.

For more information about WSGI, see PEP 333:
    http://www.python.org/dev/peps/pep-0333/
"""

from io import BytesIO

from ...bzr.smart import medium
from ...transport import chroot, get_transport
from ...urlutils import local_path_to_url


def make_app(root, prefix, path_var='REQUEST_URI', readonly=True,
             load_plugins=True, enable_logging=True):
    """Convenience function to construct a WSGI bzr smart server.

    :param root: a local path that requests will be relative to.
    :param prefix: See RelpathSetter.
    :param path_var: See RelpathSetter.
    """
    local_url = local_path_to_url(root)
    if readonly:
        base_transport = get_transport('readonly+' + local_url)
    else:
        base_transport = get_transport(local_url)
    if load_plugins:
        from ...plugin import load_plugins
        load_plugins()
    if enable_logging:
        import breezy.trace
        breezy.trace.enable_default_logging()
    app = SmartWSGIApp(base_transport, prefix)
    app = RelpathSetter(app, '', path_var)
    return app


class RelpathSetter:
    """WSGI middleware to set 'breezy.relpath' in the environ.

    Different servers can invoke a SmartWSGIApp in different ways.  This
    middleware allows an adminstrator to configure how to the SmartWSGIApp will
    determine what path it should be serving for a given request for many common
    situations.

    For example, a request for "/some/prefix/repo/branch/.bzr/smart" received by
    a typical Apache and mod_fastcgi configuration will set `REQUEST_URI` to
    "/some/prefix/repo/branch/.bzr/smart".  A RelpathSetter with
    prefix="/some/prefix/" and path_var="REQUEST_URI" will set that request's
    'breezy.relpath' variable to "repo/branch".
    """

    def __init__(self, app, prefix='', path_var='REQUEST_URI'):
        """Constructor.

        :param app: WSGI app to wrap, e.g. a SmartWSGIApp instance.
        :param path_var: the variable in the WSGI environ to calculate the
            'breezy.relpath' variable from.
        :param prefix: a prefix to strip from the variable specified in
            path_var before setting 'breezy.relpath'.
        """
        self.app = app
        self.prefix = prefix
        self.path_var = path_var

    def __call__(self, environ, start_response):
        path = environ[self.path_var]
        suffix = '/.bzr/smart'
        if not (path.startswith(self.prefix) and path.endswith(suffix)):
            start_response('404 Not Found', [])
            return []
        environ['breezy.relpath'] = path[len(self.prefix):-len(suffix)]
        return self.app(environ, start_response)


class SmartWSGIApp:
    """A WSGI application for the bzr smart server."""

    def __init__(self, backing_transport, root_client_path='/'):
        """Constructor.

        :param backing_transport: a transport.  Requests will be processed
            relative to this transport.
        :param root_client_path: the client path that maps to the root of
            backing_transport.  This is used to interpret relpaths received from
            the client.
        """
        # Use a ChrootServer so that this web application won't
        # accidentally let people access locations they shouldn't.
        # e.g. consider a smart server request for "get /etc/passwd" or
        # something.
        self.chroot_server = chroot.ChrootServer(backing_transport)
        self.chroot_server.start_server()
        self.backing_transport = get_transport(self.chroot_server.get_url())
        self.root_client_path = root_client_path
        # While the chroot server can technically be torn down at this point,
        # as all it does is remove the scheme registration from transport's
        # protocol dictionary, we don't *just in case* there are parts of
        # breezy that will invoke 'get_transport' on urls rather than cloning
        # around the existing transport.
        # self.chroot_server.stop_server()

    def __call__(self, environ, start_response):
        """WSGI application callable."""
        if environ['REQUEST_METHOD'] != 'POST':
            start_response('405 Method not allowed', [('Allow', 'POST')])
            return []

        relpath = environ['breezy.relpath']

        if not relpath.startswith('/'):
            relpath = '/' + relpath
        if not relpath.endswith('/'):
            relpath += '/'

        # Compare the HTTP path (relpath) and root_client_path, and calculate
        # new relpath and root_client_path accordingly, to be used to build the
        # request.
        if relpath.startswith(self.root_client_path):
            # The relpath traverses all of the mandatory root client path.
            # Remove the root_client_path from the relpath, and set
            # adjusted_rcp to None to tell the request handler that no further
            # path translation is required.
            adjusted_rcp = None
            adjusted_relpath = relpath[len(self.root_client_path):]
        elif self.root_client_path.startswith(relpath):
            # The relpath traverses some of the mandatory root client path.
            # Subtract the relpath from the root_client_path, and set the
            # relpath to '.'.
            adjusted_rcp = '/' + self.root_client_path[len(relpath):]
            adjusted_relpath = '.'
        else:
            adjusted_rcp = self.root_client_path
            adjusted_relpath = relpath

        if adjusted_relpath.startswith('/'):
            adjusted_relpath = adjusted_relpath[1:]
        if adjusted_relpath.startswith('/'):
            raise AssertionError(adjusted_relpath)

        transport = self.backing_transport.clone(adjusted_relpath)
        out_buffer = BytesIO()
        request_data_length = int(environ['CONTENT_LENGTH'])
        request_data_bytes = environ['wsgi.input'].read(request_data_length)
        smart_protocol_request = self.make_request(
            transport, out_buffer.write, request_data_bytes,
            adjusted_rcp)
        if smart_protocol_request.next_read_size() != 0:
            # The request appears to be incomplete, or perhaps it's just a
            # newer version we don't understand.  Regardless, all we can do
            # is return an error response in the format of our version of the
            # protocol.
            response_data = b'error\x01incomplete request\n'
        else:
            response_data = out_buffer.getvalue()
        headers = [('Content-type', 'application/octet-stream')]
        headers.append(("Content-Length", str(len(response_data))))
        start_response('200 OK', headers)
        return [response_data]

    def make_request(self, transport, write_func, request_bytes, rcp):
        protocol_factory, unused_bytes = medium._get_protocol_factory_for_bytes(
            request_bytes)
        server_protocol = protocol_factory(
            transport, write_func, rcp, self.backing_transport)
        server_protocol.accept_bytes(unused_bytes)
        return server_protocol



# Copyright (C) 2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for test feature dependencies."""

import sys

from .. import plugin as _mod_plugin
from .. import symbol_versioning, tests
from . import features


class TestFeature(tests.TestCase):

    def test_caching(self):
        """Feature._probe is called by the feature at most once."""
        class InstrumentedFeature(features.Feature):
            def __init__(self):
                super().__init__()
                self.calls = []

            def _probe(self):
                self.calls.append('_probe')
                return False
        feature = InstrumentedFeature()
        feature.available()
        self.assertEqual(['_probe'], feature.calls)
        feature.available()
        self.assertEqual(['_probe'], feature.calls)

    def test_named_str(self):
        """Feature.__str__ should thunk to feature_name()."""
        class NamedFeature(features.Feature):
            def feature_name(self):
                return 'symlinks'
        feature = NamedFeature()
        self.assertEqual('symlinks', str(feature))

    def test_default_str(self):
        """Feature.__str__ should default to __class__.__name__."""
        class NamedFeature(features.Feature):
            pass
        feature = NamedFeature()
        self.assertEqual('NamedFeature', str(feature))


class TestUnavailableFeature(tests.TestCase):

    def test_access_feature(self):
        feature = features.Feature()
        exception = tests.UnavailableFeature(feature)
        self.assertIs(feature, exception.args[0])


# Although this was deprecated a long time ago, please keep it here because
# it's really just a test fixture for test-feature deprecation.
simple_thunk_feature = features._CompatabilityThunkFeature(
    symbol_versioning.deprecated_in((2, 1, 0)),
    'breezy.tests.test_features',
    'simple_thunk_feature',
    'UnicodeFilenameFeature',
    replacement_module='breezy.tests.features')


class Test_CompatibilityFeature(tests.TestCase):

    def test_does_thunk(self):
        res = self.callDeprecated(
            ['breezy.tests.test_features.simple_thunk_feature '
             'was deprecated in version 2.1.0. '
             'Use breezy.tests.features.UnicodeFilenameFeature instead.'],
            simple_thunk_feature.available)
        self.assertEqual(features.UnicodeFilenameFeature.available(), res)

    def test_reports_correct_location(self):
        a_feature = features._CompatabilityThunkFeature(
            symbol_versioning.deprecated_in((2, 1, 0)),
            'breezy.tests.test_features',
            'a_feature',
            'UnicodeFilenameFeature',
            replacement_module='breezy.tests.features')

        def test_caller(message, category=None, stacklevel=1):
            # Find ourselves back from the right frame
            caller = sys._getframe(stacklevel)
            reported_file = caller.f_globals['__file__']
            reported_lineno = caller.f_lineno
            self.assertEqual(__file__, reported_file)
            # The call we're tracking occurred the line after we grabbed the
            # lineno.
            self.assertEqual(self.lineno + 1, reported_lineno)
        self.overrideAttr(symbol_versioning, 'warn', test_caller)
        # Grab the current lineno
        self.lineno = sys._getframe().f_lineno
        self.requireFeature(a_feature)


class TestModuleAvailableFeature(tests.TestCase):

    def test_available_module(self):
        feature = features.ModuleAvailableFeature('breezy.tests')
        self.assertEqual('breezy.tests', feature.module_name)
        self.assertEqual('breezy.tests', str(feature))
        self.assertTrue(feature.available())
        self.assertIs(tests, feature.module)

    def test_unavailable_module(self):
        feature = features.ModuleAvailableFeature(
            'breezy.no_such_module_exists')
        self.assertEqual('breezy.no_such_module_exists', str(feature))
        self.assertFalse(feature.available())
        self.assertIs(None, feature.module)


class TestPluginLoadedFeature(tests.TestCase):

    def test_available_plugin(self):
        plugins = _mod_plugin.plugins()
        if not plugins:
            self.skipTest('no plugins available to test with')
        a_plugin_name = next(iter(plugins))
        feature = features.PluginLoadedFeature(a_plugin_name)
        self.assertEqual(a_plugin_name, feature.plugin_name)
        self.assertEqual(a_plugin_name + ' plugin', str(feature))
        self.assertTrue(feature.available())

    def test_unavailable_plugin(self):
        feature = features.PluginLoadedFeature('idontexist')
        self.assertEqual('idontexist plugin', str(feature))
        self.assertFalse(feature.available())
        self.assertIs(None, feature.plugin)


class TestUnicodeFilenameFeature(tests.TestCase):

    def test_probe_passes(self):
        """UnicodeFilenameFeature._probe passes."""
        # We can't test much more than that because the behaviour depends
        # on the platform.
        features.UnicodeFilenameFeature._probe()


class TestBackslashFilenameFeature(tests.TestCase):

    def test_probe_passes(self):
        """BackslashFilenameFeature._probe passes."""
        # We can't test much more than that because the behaviour depends
        # on the platform.
        features.BackslashFilenameFeature._probe()



# Copyright (C) 2007, 2009-2012 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

# Original author: David Allouche

from . import errors, lock, merge, revision
from .branch import Branch
from .i18n import gettext
from .trace import note


def _run_post_switch_hooks(control_dir, to_branch, force, revision_id):
    from .branch import SwitchHookParams
    hooks = Branch.hooks['post_switch']
    if not hooks:
        return
    params = SwitchHookParams(control_dir, to_branch, force, revision_id)
    for hook in hooks:
        hook(params)


def switch(control_dir, to_branch, force=False, quiet=False, revision_id=None,
           store_uncommitted=False, possible_transports=None):
    """Switch the branch associated with a checkout.

    :param control_dir: ControlDir of the checkout to change
    :param to_branch: branch that the checkout is to reference
    :param force: skip the check for local commits in a heavy checkout
    :param revision_id: revision ID to switch to.
    :param store_uncommitted: If True, store uncommitted changes in the
        branch.
    """
    try:
        tree = control_dir.open_workingtree()
    except errors.NotBranchError as ex:
        # Lightweight checkout and branch is no longer there
        if not force or store_uncommitted:
            raise ex
        else:
            tree = None
    else:
        if store_uncommitted or tree.branch.get_bound_location():
            tree.lock_write()
        else:
            tree.lock_tree_write()
    try:
        if tree is not None:
            parent_ids = tree.get_parent_ids()
            if len(parent_ids) > 1:
                raise errors.CommandError(
                    gettext('Pending merges must be '
                            'committed or reverted before using switch.'))

        if store_uncommitted:
            tree.store_uncommitted()
        if tree is None:
            source_repository = to_branch.repository
            base_revision_id = None
        else:
            source_repository = tree.branch.repository
            # Attempt to retrieve the base_revision_id now, since some
            # working tree formats (i.e. git) don't have their own
            # last_revision but just use that of the currently active branch.
            base_revision_id = tree.last_revision()
    finally:
        if tree is not None:
            tree.unlock()
    with to_branch.lock_read():
        _set_branch_location(control_dir, to_branch, tree.branch if tree else None, force)
    tree = control_dir.open_workingtree()
    if store_uncommitted:
        tree.lock_write()
    else:
        tree.lock_tree_write()
    try:
        if base_revision_id is None:
            # If we couldn't get to the tree's last_revision earlier, perhaps
            # we can now.
            base_revision_id = tree.last_revision()
        if revision_id is None:
            revision_id = to_branch.last_revision()
        if base_revision_id == revision_id:
            if not quiet:
                note(gettext("Tree is up to date at revision %d."),
                     to_branch.revno())
        else:
            base_tree = source_repository.revision_tree(base_revision_id)
            target_tree = to_branch.repository.revision_tree(revision_id)
            merge.Merge3Merger(tree, tree, base_tree, target_tree)
            tree.set_last_revision(revision_id)
            if not quiet:
                note(gettext('Updated to revision %d.') % to_branch.revno())
        if store_uncommitted:
            tree.restore_uncommitted()
        _run_post_switch_hooks(control_dir, to_branch, force, revision_id)
    finally:
        tree.unlock()


def _set_branch_location(control, to_branch, current_branch, force=False):
    """Set location value of a branch reference.

    :param control: ControlDir of the checkout to change
    :param to_branch: branch that the checkout is to reference
    :param force: skip the check for local commits in a heavy checkout
    """
    branch_format = control.find_branch_format()
    if branch_format.get_reference(control) is not None:
        # Lightweight checkout: update the branch reference
        branch_format.set_reference(control, None, to_branch)
    else:
        b = current_branch
        bound_branch = b.get_bound_location()
        if bound_branch is not None:
            # Heavyweight checkout: check all local commits
            # have been pushed to the current bound branch then
            # synchronise the local branch with the new remote branch
            # and bind to it
            possible_transports = []
            try:
                if not force and _any_local_commits(b, possible_transports):
                    raise errors.CommandError(gettext(
                        'Cannot switch as local commits found in the checkout. '
                        'Commit these to the bound branch or use --force to '
                        'throw them away.'))
            except errors.BoundBranchConnectionFailure as e:
                raise errors.CommandError(gettext(
                    'Unable to connect to current master branch %(target)s: '
                    '%(error)s To switch anyway, use --force.') %
                    e.__dict__)
            with b.lock_write():
                b.set_bound_location(None)
                b.pull(to_branch, overwrite=True,
                       possible_transports=possible_transports)
                b.set_bound_location(to_branch.base)
                b.set_parent(b.get_master_branch().get_parent())
        else:
            # If this is a standalone tree and the new branch
            # is derived from this one, create a lightweight checkout.
            with b.lock_read():
                graph = b.repository.get_graph(to_branch.repository)
                if (b.controldir._format.colocated_branches and
                    (force or graph.is_ancestor(
                        b.last_revision(), to_branch.last_revision()))):
                    b.controldir.destroy_branch()
                    b.controldir.set_branch_reference(to_branch, name="")
                else:
                    raise errors.CommandError(
                        gettext('Cannot switch a branch, only a checkout.'))


def _any_local_commits(this_branch, possible_transports):
    """Does this branch have any commits not in the master branch?"""
    last_rev = this_branch.last_revision()
    if last_rev != revision.NULL_REVISION:
        other_branch = this_branch.get_master_branch(possible_transports)
        with this_branch.lock_read(), other_branch.lock_read():
            other_last_rev = other_branch.last_revision()
            graph = this_branch.repository.get_graph(
                other_branch.repository)
            if not graph.is_ancestor(last_rev, other_last_rev):
                return True
    return False



# Copyright (C) 2009-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from ... import errors
from ...bzr.vf_search import PendingAncestryResult
from ...commands import Command
from ...controldir import ControlDir
from ...option import Option
from ...repository import WriteGroup
from ...revision import NULL_REVISION


class cmd_fix_missing_keys_for_stacking(Command):
    """Fix missing keys for stacking.

    This is the fixer script for <https://bugs.launchpad.net/bzr/+bug/354036>.
    """

    hidden = True
    takes_args = ['branch_url']
    takes_options = [
        Option(
            'dry-run',
            help="Show what would be done, but don't actually do anything."),
        ]

    def run(self, branch_url, dry_run=False):
        try:
            bd = ControlDir.open(branch_url)
            b = bd.open_branch(ignore_fallbacks=True)
        except (errors.NotBranchError, errors.InvalidURL):
            raise errors.CommandError(
                "Not a branch or invalid URL: %s" % branch_url)
        b.lock_read()
        try:
            url = b.get_stacked_on_url()
        except (errors.UnstackableRepositoryFormat, errors.NotStacked,
                errors.UnstackableBranchFormat):
            b.unlock()
            raise errors.CommandError("Not stacked: %s" % branch_url)
        raw_r = b.repository.controldir.open_repository()
        if dry_run:
            raw_r.lock_read()
        else:
            b.unlock()
            b = b.controldir.open_branch()
            b.lock_read()
            raw_r.lock_write()
        try:
            revs = raw_r.all_revision_ids()
            rev_parents = raw_r.get_graph().get_parent_map(revs)
            needed = set()
            map(needed.update, rev_parents.values())
            needed.discard(NULL_REVISION)
            needed = {(rev,) for rev in needed}
            needed = needed - raw_r.inventories.keys()
            if not needed:
                # Nothing to see here.
                return
            self.outf.write("Missing inventories: %r\n" % needed)
            if dry_run:
                return
            assert raw_r._format.network_name() == b.repository._format.network_name()
            stream = b.repository.inventories.get_record_stream(
                needed, 'topological', True)
            with WriteGroup(raw_r):
                raw_r.inventories.insert_record_stream(stream)
        finally:
            raw_r.unlock()
        b.unlock()
        self.outf.write("Fixed: %s\n" % branch_url)


class cmd_mirror_revs_into(Command):
    """Mirror all revs from one repo into another."""

    hidden = True
    takes_args = ['source', 'destination']

    _see_also = ['fetch-all-records']

    def run(self, source, destination):
        bd = ControlDir.open(source)
        source_r = bd.open_branch().repository
        bd = ControlDir.open(destination)
        target_r = bd.open_branch().repository
        with source_r.lock_read(), target_r.lock_write():
            revs = [k[-1] for k in source_r.revisions.keys()]
            target_r.fetch(
                source_r, fetch_spec=PendingAncestryResult(revs, source_r))



# Copyright (C) 2005, 2006, 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for different inventory implementations"""

# NOTE: Don't import Inventory here, to make sure that we don't accidentally
# hardcode that when we should be using self.make_inventory

from breezy import errors, osutils
from breezy.bzr import inventory
from breezy.bzr.inventory import (InventoryDirectory, InventoryEntry,
                                  InventoryFile, InventoryLink, TreeReference)
from breezy.bzr.tests.per_inventory import TestCaseWithInventory


class TestInventory(TestCaseWithInventory):

    def make_init_inventory(self):
        inv = inventory.Inventory(b'tree-root')
        inv.revision = b'initial-rev'
        inv.root.revision = b'initial-rev'
        return self.inv_to_test_inv(inv)

    def make_file(self, file_id, name, parent_id, content=b'content\n',
                  revision=b'new-test-rev'):
        ie = InventoryFile(file_id, name, parent_id)
        ie.text_sha1 = osutils.sha_string(content)
        ie.text_size = len(content)
        ie.revision = revision
        return ie

    def make_link(self, file_id, name, parent_id, target='link-target\n'):
        ie = InventoryLink(file_id, name, parent_id)
        ie.symlink_target = target
        return ie

    def prepare_inv_with_nested_dirs(self):
        inv = inventory.Inventory(b'tree-root')
        inv.root.revision = b'revision'
        for args in [('src', 'directory', b'src-id'),
                     ('doc', 'directory', b'doc-id'),
                     ('src/hello.c', 'file', b'hello-id'),
                     ('src/bye.c', 'file', b'bye-id'),
                     ('zz', 'file', b'zz-id'),
                     ('src/sub/', 'directory', b'sub-id'),
                     ('src/zz.c', 'file', b'zzc-id'),
                     ('src/sub/a', 'file', b'a-id'),
                     ('Makefile', 'file', b'makefile-id')]:
            ie = inv.add_path(*args)
            ie.revision = b'revision'
            if args[1] == 'file':
                ie.text_sha1 = osutils.sha_string(b'content\n')
                ie.text_size = len(b'content\n')
        return self.inv_to_test_inv(inv)


class TestInventoryCreateByApplyDelta(TestInventory):
    """A subset of the inventory delta application tests.

    See test_inv which has comprehensive delta application tests for
    inventories, dirstate, and repository based inventories.
    """

    def test_add(self):
        inv = self.make_init_inventory()
        inv = inv.create_by_apply_delta([
            (None, "a", b"a-id", self.make_file(b'a-id', 'a', b'tree-root')),
            ], b'new-test-rev')
        self.assertEqual('a', inv.id2path(b'a-id'))

    def test_delete(self):
        inv = self.make_init_inventory()
        inv = inv.create_by_apply_delta([
            (None, "a", b"a-id", self.make_file(b'a-id', 'a', b'tree-root')),
            ], b'new-rev-1')
        self.assertEqual('a', inv.id2path(b'a-id'))
        inv = inv.create_by_apply_delta([
            ("a", None, b"a-id", None),
            ], b'new-rev-2')
        self.assertRaises(errors.NoSuchId, inv.id2path, b'a-id')

    def test_rename(self):
        inv = self.make_init_inventory()
        inv = inv.create_by_apply_delta([
            (None, "a", b"a-id", self.make_file(b'a-id', 'a', b'tree-root')),
            ], b'new-rev-1')
        self.assertEqual('a', inv.id2path(b'a-id'))
        a_ie = inv.get_entry(b'a-id')
        b_ie = self.make_file(a_ie.file_id, "b", a_ie.parent_id)
        inv = inv.create_by_apply_delta(
            [("a", "b", b"a-id", b_ie)], b'new-rev-2')
        self.assertEqual("b", inv.id2path(b'a-id'))

    def test_illegal(self):
        # A file-id cannot appear in a delta more than once
        inv = self.make_init_inventory()
        self.assertRaises(errors.InconsistentDelta, inv.create_by_apply_delta, [
            (None, "a", b"id-1", self.make_file(b'id-1', 'a', b'tree-root')),
            (None, "b", b"id-1", self.make_file(b'id-1', 'b', b'tree-root')),
            ], b'new-rev-1')


class TestInventoryReads(TestInventory):

    def test_is_root(self):
        """Ensure our root-checking code is accurate."""
        inv = self.make_init_inventory()
        self.assertTrue(inv.is_root(b'tree-root'))
        self.assertFalse(inv.is_root(b'booga'))
        ie = inv.get_entry(b'tree-root').copy()
        ie.file_id = b'booga'
        inv = inv.create_by_apply_delta([("", None, b"tree-root", None),
                                         (None, "", b"booga", ie)], b'new-rev-2')
        self.assertFalse(inv.is_root(b'TREE_ROOT'))
        self.assertTrue(inv.is_root(b'booga'))

    def test_ids(self):
        """Test detection of files within selected directories."""
        inv = inventory.Inventory(b'TREE_ROOT')
        inv.root.revision = b'revision'
        for args in [('src', 'directory', b'src-id'),
                     ('doc', 'directory', b'doc-id'),
                     ('src/hello.c', 'file'),
                     ('src/bye.c', 'file', b'bye-id'),
                     ('Makefile', 'file')]:
            ie = inv.add_path(*args)
            ie.revision = b'revision'
            if args[1] == 'file':
                ie.text_sha1 = osutils.sha_string(b'content\n')
                ie.text_size = len(b'content\n')
        inv = self.inv_to_test_inv(inv)
        self.assertEqual(inv.path2id('src'), b'src-id')
        self.assertEqual(inv.path2id('src/bye.c'), b'bye-id')

    def test_get_entry_by_path_partial(self):
        inv = inventory.Inventory(b'TREE_ROOT')
        inv.root.revision = b'revision'
        for args in [('src', 'directory', b'src-id'),
                     ('doc', 'directory', b'doc-id'),
                     ('src/hello.c', 'file'),
                     ('src/bye.c', 'file', b'bye-id'),
                     ('Makefile', 'file'),
                     ('external', 'tree-reference', b'other-root')]:
            ie = inv.add_path(*args)
            ie.revision = b'revision'
            if args[1] == 'file':
                ie.text_sha1 = osutils.sha_string(b'content\n')
                ie.text_size = len(b'content\n')
            if args[1] == 'tree-reference':
                ie.reference_revision = b'reference'
        inv = self.inv_to_test_inv(inv)

        # Standard lookups
        ie, resolved, remaining = inv.get_entry_by_path_partial('')
        self.assertEqual((ie.file_id, resolved, remaining), (b'TREE_ROOT', [], []))
        ie, resolved, remaining = inv.get_entry_by_path_partial('src')
        self.assertEqual((ie.file_id, resolved, remaining), (b'src-id', ['src'], []))
        ie, resolved, remaining = inv.get_entry_by_path_partial('src/bye.c')
        self.assertEqual((ie.file_id, resolved, remaining), (b'bye-id', ['src', 'bye.c'], []))

        # Paths in the external tree
        ie, resolved, remaining = inv.get_entry_by_path_partial('external')
        self.assertEqual((ie.file_id, resolved, remaining), (b'other-root', ['external'], []))
        ie, resolved, remaining = inv.get_entry_by_path_partial('external/blah')
        self.assertEqual((ie.file_id, resolved, remaining), (b'other-root', ['external'], ['blah']))

        # Nonexistant paths
        ie, resolved, remaining = inv.get_entry_by_path_partial('foo.c')
        self.assertEqual((ie, resolved, remaining), (None, None, None))

    def test_non_directory_children(self):
        """Test path2id when a parent directory has no children"""
        inv = inventory.Inventory(b'tree-root')
        inv.add(self.make_file(b'file-id', 'file', b'tree-root'))
        inv.add(self.make_link(b'link-id', 'link', b'tree-root'))
        self.assertIs(None, inv.path2id('file/subfile'))
        self.assertIs(None, inv.path2id('link/subfile'))

    def test_is_unmodified(self):
        f1 = self.make_file(b'file-id', 'file', b'tree-root')
        f1.revision = b'rev'
        self.assertTrue(f1.is_unmodified(f1))
        f2 = self.make_file(b'file-id', 'file', b'tree-root')
        f2.revision = b'rev'
        self.assertTrue(f1.is_unmodified(f2))
        f3 = self.make_file(b'file-id', 'file', b'tree-root')
        self.assertFalse(f1.is_unmodified(f3))
        f4 = self.make_file(b'file-id', 'file', b'tree-root')
        f4.revision = b'rev1'
        self.assertFalse(f1.is_unmodified(f4))

    def test_iter_entries(self):
        inv = self.prepare_inv_with_nested_dirs()

        # Test all entries
        self.assertEqual([
            ('', b'tree-root'),
            ('Makefile', b'makefile-id'),
            ('doc', b'doc-id'),
            ('src', b'src-id'),
            ('src/bye.c', b'bye-id'),
            ('src/hello.c', b'hello-id'),
            ('src/sub', b'sub-id'),
            ('src/sub/a', b'a-id'),
            ('src/zz.c', b'zzc-id'),
            ('zz', b'zz-id'),
            ], [(path, ie.file_id) for path, ie in inv.iter_entries()])

        # Test a subdirectory
        self.assertEqual([
            ('bye.c', b'bye-id'),
            ('hello.c', b'hello-id'),
            ('sub', b'sub-id'),
            ('sub/a', b'a-id'),
            ('zz.c', b'zzc-id'),
            ], [(path, ie.file_id) for path, ie in inv.iter_entries(
                from_dir=b'src-id')])

        # Test not recursing at the root level
        self.assertEqual([
            ('', b'tree-root'),
            ('Makefile', b'makefile-id'),
            ('doc', b'doc-id'),
            ('src', b'src-id'),
            ('zz', b'zz-id'),
            ], [(path, ie.file_id) for path, ie in inv.iter_entries(
                recursive=False)])

        # Test not recursing at a subdirectory level
        self.assertEqual([
            ('bye.c', b'bye-id'),
            ('hello.c', b'hello-id'),
            ('sub', b'sub-id'),
            ('zz.c', b'zzc-id'),
            ], [(path, ie.file_id) for path, ie in inv.iter_entries(
                from_dir=b'src-id', recursive=False)])

    def test_iter_just_entries(self):
        inv = self.prepare_inv_with_nested_dirs()
        self.assertEqual([
            b'a-id',
            b'bye-id',
            b'doc-id',
            b'hello-id',
            b'makefile-id',
            b'src-id',
            b'sub-id',
            b'tree-root',
            b'zz-id',
            b'zzc-id',
            ], sorted([ie.file_id for ie in inv.iter_just_entries()]))

    def test_iter_entries_by_dir(self):
        inv = self. prepare_inv_with_nested_dirs()
        self.assertEqual([
            ('', b'tree-root'),
            ('Makefile', b'makefile-id'),
            ('doc', b'doc-id'),
            ('src', b'src-id'),
            ('zz', b'zz-id'),
            ('src/bye.c', b'bye-id'),
            ('src/hello.c', b'hello-id'),
            ('src/sub', b'sub-id'),
            ('src/zz.c', b'zzc-id'),
            ('src/sub/a', b'a-id'),
            ], [(path, ie.file_id) for path, ie in inv.iter_entries_by_dir()])
        self.assertEqual([
            ('', b'tree-root'),
            ('Makefile', b'makefile-id'),
            ('doc', b'doc-id'),
            ('src', b'src-id'),
            ('zz', b'zz-id'),
            ('src/bye.c', b'bye-id'),
            ('src/hello.c', b'hello-id'),
            ('src/sub', b'sub-id'),
            ('src/zz.c', b'zzc-id'),
            ('src/sub/a', b'a-id'),
            ], [(path, ie.file_id) for path, ie in inv.iter_entries_by_dir(
                specific_file_ids=(b'a-id', b'zzc-id', b'doc-id', b'tree-root',
                                   b'hello-id', b'bye-id', b'zz-id', b'src-id', b'makefile-id',
                                   b'sub-id'))])

        self.assertEqual([
            ('Makefile', b'makefile-id'),
            ('doc', b'doc-id'),
            ('zz', b'zz-id'),
            ('src/bye.c', b'bye-id'),
            ('src/hello.c', b'hello-id'),
            ('src/zz.c', b'zzc-id'),
            ('src/sub/a', b'a-id'),
            ], [(path, ie.file_id) for path, ie in inv.iter_entries_by_dir(
                specific_file_ids=(b'a-id', b'zzc-id', b'doc-id',
                                   b'hello-id', b'bye-id', b'zz-id', b'makefile-id'))])

        self.assertEqual([
            ('Makefile', b'makefile-id'),
            ('src/bye.c', b'bye-id'),
            ], [(path, ie.file_id) for path, ie in inv.iter_entries_by_dir(
                specific_file_ids=(b'bye-id', b'makefile-id'))])

        self.assertEqual([
            ('Makefile', b'makefile-id'),
            ('src/bye.c', b'bye-id'),
            ], [(path, ie.file_id) for path, ie in inv.iter_entries_by_dir(
                specific_file_ids=(b'bye-id', b'makefile-id'))])

        self.assertEqual([
            ('src/bye.c', b'bye-id'),
            ], [(path, ie.file_id) for path, ie in inv.iter_entries_by_dir(
                specific_file_ids=(b'bye-id',))])


class TestInventoryFiltering(TestInventory):

    def test_inv_filter_empty(self):
        inv = self.prepare_inv_with_nested_dirs()
        new_inv = inv.filter([])
        self.assertEqual([
            ('', b'tree-root'),
            ], [(path, ie.file_id) for path, ie in new_inv.iter_entries()])

    def test_inv_filter_files(self):
        inv = self.prepare_inv_with_nested_dirs()
        new_inv = inv.filter([b'zz-id', b'hello-id', b'a-id'])
        self.assertEqual([
            ('', b'tree-root'),
            ('src', b'src-id'),
            ('src/hello.c', b'hello-id'),
            ('src/sub', b'sub-id'),
            ('src/sub/a', b'a-id'),
            ('zz', b'zz-id'),
            ], [(path, ie.file_id) for path, ie in new_inv.iter_entries()])

    def test_inv_filter_dirs(self):
        inv = self.prepare_inv_with_nested_dirs()
        new_inv = inv.filter([b'doc-id', b'sub-id'])
        self.assertEqual([
            ('', b'tree-root'),
            ('doc', b'doc-id'),
            ('src', b'src-id'),
            ('src/sub', b'sub-id'),
            ('src/sub/a', b'a-id'),
            ], [(path, ie.file_id) for path, ie in new_inv.iter_entries()])

    def test_inv_filter_files_and_dirs(self):
        inv = self.prepare_inv_with_nested_dirs()
        new_inv = inv.filter([b'makefile-id', b'src-id'])
        self.assertEqual([
            ('', b'tree-root'),
            ('Makefile', b'makefile-id'),
            ('src', b'src-id'),
            ('src/bye.c', b'bye-id'),
            ('src/hello.c', b'hello-id'),
            ('src/sub', b'sub-id'),
            ('src/sub/a', b'a-id'),
            ('src/zz.c', b'zzc-id'),
            ], [(path, ie.file_id) for path, ie in new_inv.iter_entries()])

    def test_inv_filter_entry_not_present(self):
        inv = self.prepare_inv_with_nested_dirs()
        new_inv = inv.filter([b'not-present-id'])
        self.assertEqual([
            ('', b'tree-root'),
            ], [(path, ie.file_id) for path, ie in new_inv.iter_entries()])



# Copyright (C) 2008, 2009, 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Tests for upgrades of various stacking situations."""

from .. import check, controldir, errors, tests
from ..upgrade import upgrade
from .scenarios import load_tests_apply_scenarios


def upgrade_scenarios():
    scenario_pairs = [  # old format, new format, model_change
        #        ('knit', 'rich-root', True),
        ('knit', '1.6', False),
        #        ('pack-0.92', '1.6', False),
        ('1.6', '1.6.1-rich-root', True),
        ]
    scenarios = []
    for (old_name, new_name, model_change) in scenario_pairs:
        name = old_name + ', ' + new_name
        scenarios.append((name,
                          dict(scenario_old_format=old_name,
                               scenario_new_format=new_name,
                               scenario_model_change=model_change)))
    return scenarios


load_tests = load_tests_apply_scenarios


class TestStackUpgrade(tests.TestCaseWithTransport):
    # TODO: This should possibly be repeated for all stacking repositories,
    # pairwise by rich/non-rich format; should possibly also try other kinds
    # of upgrades like knit->pack. -- mbp 20080804

    scenarios = upgrade_scenarios()

    def test_stack_upgrade(self):
        """Correct checks when stacked-on repository is upgraded.

        We initially stack on a repo with the same rich root support,
        we then upgrade it and should fail, we then upgrade the overlaid
        repository.
        """
        base = self.make_branch_and_tree('base',
                                         format=self.scenario_old_format)
        self.build_tree(['base/foo'])
        base.commit('base commit')
        # make another one stacked
        stacked = base.controldir.sprout('stacked', stacked=True)
        # this must really be stacked (or get_stacked_on_url raises an error)
        self.assertTrue(stacked.open_branch().get_stacked_on_url())
        # now we'll upgrade the underlying branch, then upgrade the stacked
        # branch, and this should still work.
        new_format = controldir.format_registry.make_controldir(
            self.scenario_new_format)
        upgrade('base', new_format)
        # in some cases you'll get an error if the underlying model has
        # changed; if just the data format has changed this should still work
        if self.scenario_model_change:
            self.assertRaises(errors.IncompatibleRepositories,
                              stacked.open_branch)
        else:
            check.check_dwim('stacked', False, True, True)
        stacked = controldir.ControlDir.open('stacked')
        # but we can upgrade the stacked repository
        upgrade('stacked', new_format)
        # and now it opens ok
        stacked = controldir.ControlDir.open('stacked')
        # And passes check.
        check.check_dwim('stacked', False, True, True)



# Copyright (C) 2005-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""XML externalization support."""

# "XML is like violence: if it doesn't solve your problem, you aren't
# using enough of it." -- various

# importing this module is fairly slow because it has to load several
# ElementTree bits

import re
from typing import Dict, Union
from xml.etree.ElementTree import (Element, ElementTree, ParseError,
                                   SubElement, fromstring, fromstringlist,
                                   tostring, tostringlist)

from .. import errors, lazy_regex
from . import inventory, serializer


class XMLSerializer(serializer.Serializer):
    """Abstract XML object serialize/deserialize"""

    squashes_xml_invalid_characters = True

    def read_inventory_from_lines(self, lines, revision_id=None,
                                  entry_cache=None, return_from_cache=False):
        """Read xml_string into an inventory object.

        :param chunks: The xml to read.
        :param revision_id: If not-None, the expected revision id of the
            inventory. Some serialisers use this to set the results' root
            revision. This should be supplied for deserialising all
            from-repository inventories so that xml5 inventories that were
            serialised without a revision identifier can be given the right
            revision id (but not for working tree inventories where users can
            edit the data without triggering checksum errors or anything).
        :param entry_cache: An optional cache of InventoryEntry objects. If
            supplied we will look up entries via (file_id, revision_id) which
            should map to a valid InventoryEntry (File/Directory/etc) object.
        :param return_from_cache: Return entries directly from the cache,
            rather than copying them first. This is only safe if the caller
            promises not to mutate the returned inventory entries, but it can
            make some operations significantly faster.
        """
        try:
            return self._unpack_inventory(fromstringlist(lines), revision_id,
                                          entry_cache=entry_cache,
                                          return_from_cache=return_from_cache)
        except ParseError as e:
            raise serializer.UnexpectedInventoryFormat(str(e))

    def read_inventory(self, f, revision_id=None):
        try:
            try:
                return self._unpack_inventory(self._read_element(f),
                                              revision_id=None)
            finally:
                f.close()
        except ParseError as e:
            raise serializer.UnexpectedInventoryFormat(str(e))

    def write_revision_to_string(self, rev):
        return b''.join(self.write_revision_to_lines(rev))

    def read_revision(self, f):
        return self._unpack_revision(self._read_element(f))

    def read_revision_from_string(self, xml_string):
        return self._unpack_revision(fromstring(xml_string))

    def _read_element(self, f):
        return ElementTree().parse(f)


def escape_invalid_chars(message):
    """Escape the XML-invalid characters in a commit message.

    :param message: Commit message to escape
    :return: tuple with escaped message and number of characters escaped
    """
    if message is None:
        return None, 0
    # Python strings can include characters that can't be
    # represented in well-formed XML; escape characters that
    # aren't listed in the XML specification
    # (http://www.w3.org/TR/REC-xml/#NT-Char).
    return re.subn('[^\x09\x0A\x0D\u0020-\uD7FF\uE000-\uFFFD]+',
                   lambda match: match.group(0).encode(
                       'unicode_escape').decode('ascii'),
                   message)


def get_utf8_or_ascii(a_str):
    """Return a cached version of the string.

    cElementTree will return a plain string if the XML is plain ascii. It only
    returns Unicode when it needs to. We want to work in utf-8 strings. So if
    cElementTree returns a plain string, we can just return the cached version.
    If it is Unicode, then we need to encode it.

    :param a_str: An 8-bit string or Unicode as returned by
                  cElementTree.Element.get()
    :return: A utf-8 encoded 8-bit string.
    """
    # This is fairly optimized because we know what cElementTree does, this is
    # not meant as a generic function for all cases. Because it is possible for
    # an 8-bit string to not be ascii or valid utf8.
    if a_str.__class__ is str:
        return a_str.encode('utf-8')
    else:
        return a_str


_utf8_re = lazy_regex.lazy_compile(b'[&<>\'\"]|[\x80-\xff]+')
_unicode_re = lazy_regex.lazy_compile('[&<>\'\"\u0080-\uffff]')


_xml_escape_map = {
    "&": '&amp;',
    "'": "&apos;",  # FIXME: overkill
    "\"": "&quot;",
    "<": "&lt;",
    ">": "&gt;",
    }


def _unicode_escape_replace(match, _map=_xml_escape_map):
    """Replace a string of non-ascii, non XML safe characters with their escape

    This will escape both Standard XML escapes, like <>"', etc.
    As well as escaping non ascii characters, because ElementTree did.
    This helps us remain compatible to older versions of bzr. We may change
    our policy in the future, though.
    """
    # jam 20060816 Benchmarks show that try/KeyError is faster if you
    # expect the entity to rarely miss. There is about a 10% difference
    # in overall time. But if you miss frequently, then if None is much
    # faster. For our use case, we *rarely* have a revision id, file id
    # or path name that is unicode. So use try/KeyError.
    try:
        return _map[match.group()]
    except KeyError:
        return "&#%d;" % ord(match.group())


def _utf8_escape_replace(match, _map=_xml_escape_map):
    """Escape utf8 characters into XML safe ones.

    This uses 2 tricks. It is either escaping "standard" characters, like "&<>,
    or it is handling characters with the high-bit set. For ascii characters,
    we just lookup the replacement in the dictionary. For everything else, we
    decode back into Unicode, and then use the XML escape code.
    """
    try:
        return _map[match.group().decode('ascii', 'replace')].encode()
    except KeyError:
        return b''.join(b'&#%d;' % ord(uni_chr)
                        for uni_chr in match.group().decode('utf8'))


_to_escaped_map: Dict[Union[bytes, str], str] = {}


def encode_and_escape(unicode_or_utf8_str, _map=_to_escaped_map):
    """Encode the string into utf8, and escape invalid XML characters"""
    # We frequently get entities we have not seen before, so it is better
    # to check if None, rather than try/KeyError
    text = _map.get(unicode_or_utf8_str)
    if text is None:
        if isinstance(unicode_or_utf8_str, str):
            # The alternative policy is to do a regular UTF8 encoding
            # and then escape only XML meta characters.
            # Performance is equivalent once you use codecs. *However*
            # this makes the serialized texts incompatible with old versions
            # of bzr. So no net gain. (Perhaps the read code would handle utf8
            # better than entity escapes, but cElementTree seems to do just
            # fine either way)
            text = _unicode_re.sub(
                _unicode_escape_replace, unicode_or_utf8_str).encode()
        else:
            # Plain strings are considered to already be in utf-8 so we do a
            # slightly different method for escaping.
            text = _utf8_re.sub(_utf8_escape_replace,
                                unicode_or_utf8_str)
        _map[unicode_or_utf8_str] = text
    return text


def _clear_cache():
    """Clean out the unicode => escaped map"""
    _to_escaped_map.clear()


def unpack_inventory_entry(elt, entry_cache=None, return_from_cache=False):
    elt_get = elt.get
    file_id = elt_get('file_id')
    revision = elt_get('revision')
    # Check and see if we have already unpacked this exact entry
    # Some timings for "repo.revision_trees(last_100_revs)"
    #               bzr     mysql
    #   unmodified  4.1s    40.8s
    #   using lru   3.5s
    #   using fifo  2.83s   29.1s
    #   lru._cache  2.8s
    #   dict        2.75s   26.8s
    #   inv.add     2.5s    26.0s
    #   no_copy     2.00s   20.5s
    #   no_c,dict   1.95s   18.0s
    # Note that a cache of 10k nodes is more than sufficient to hold all of
    # the inventory for the last 100 revs for bzr, but not for mysql (20k
    # is enough for mysql, which saves the same 2s as using a dict)

    # Breakdown of mysql using time.clock()
    #   4.1s    2 calls to element.get for file_id, revision_id
    #   4.5s    cache_hit lookup
    #   7.1s    InventoryFile.copy()
    #   2.4s    InventoryDirectory.copy()
    #   0.4s    decoding unique entries
    #   1.6s    decoding entries after FIFO fills up
    #   0.8s    Adding nodes to FIFO (including flushes)
    #   0.1s    cache miss lookups
    # Using an LRU cache
    #   4.1s    2 calls to element.get for file_id, revision_id
    #   9.9s    cache_hit lookup
    #   10.8s   InventoryEntry.copy()
    #   0.3s    cache miss lookus
    #   1.2s    decoding entries
    #   1.0s    adding nodes to LRU
    if entry_cache is not None and revision is not None:
        key = (file_id, revision)
        try:
            # We copy it, because some operations may mutate it
            cached_ie = entry_cache[key]
        except KeyError:
            pass
        else:
            # Only copying directory entries drops us 2.85s => 2.35s
            if return_from_cache:
                if cached_ie.kind == 'directory':
                    return cached_ie.copy()
                return cached_ie
            return cached_ie.copy()

    kind = elt.tag
    if not inventory.InventoryEntry.versionable_kind(kind):
        raise AssertionError('unsupported entry kind %s' % kind)

    file_id = get_utf8_or_ascii(file_id)
    if revision is not None:
        revision = get_utf8_or_ascii(revision)
    parent_id = elt_get('parent_id')
    if parent_id is not None:
        parent_id = get_utf8_or_ascii(parent_id)

    if kind == 'directory':
        ie = inventory.InventoryDirectory(file_id,
                                          elt_get('name'),
                                          parent_id)
    elif kind == 'file':
        ie = inventory.InventoryFile(file_id,
                                     elt_get('name'),
                                     parent_id)
        ie.text_sha1 = elt_get('text_sha1')
        if ie.text_sha1 is not None:
            ie.text_sha1 = ie.text_sha1.encode('ascii')
        if elt_get('executable') == 'yes':
            ie.executable = True
        v = elt_get('text_size')
        ie.text_size = v and int(v)
    elif kind == 'symlink':
        ie = inventory.InventoryLink(file_id,
                                     elt_get('name'),
                                     parent_id)
        ie.symlink_target = elt_get('symlink_target')
    elif kind == 'tree-reference':
        file_id = get_utf8_or_ascii(elt.attrib['file_id'])
        name = elt.attrib['name']
        parent_id = get_utf8_or_ascii(elt.attrib['parent_id'])
        revision = get_utf8_or_ascii(elt.get('revision'))
        reference_revision = get_utf8_or_ascii(elt.get('reference_revision'))
        ie = inventory.TreeReference(file_id, name, parent_id, revision,
                                     reference_revision)
    else:
        raise serializer.UnsupportedInventoryKind(kind)
    ie.revision = revision
    if revision is not None and entry_cache is not None:
        # We cache a copy() because callers like to mutate objects, and
        # that would cause the item in cache to mutate as well.
        # This has a small effect on many-inventory performance, because
        # the majority fraction is spent in cache hits, not misses.
        entry_cache[key] = ie.copy()

    return ie


def unpack_inventory_flat(elt, format_num, unpack_entry,
                          entry_cache=None, return_from_cache=False):
    """Unpack a flat XML inventory.

    :param elt: XML element for the inventory
    :param format_num: Expected format number
    :param unpack_entry: Function for unpacking inventory entries
    :return: An inventory
    :raise UnexpectedInventoryFormat: When unexpected elements or data is
        encountered
    """
    if elt.tag != 'inventory':
        raise serializer.UnexpectedInventoryFormat('Root tag is %r' % elt.tag)
    format = elt.get('format')
    if ((format is None and format_num is not None) or
            format.encode() != format_num):
        raise serializer.UnexpectedInventoryFormat('Invalid format version %r' % format)
    revision_id = elt.get('revision_id')
    if revision_id is not None:
        revision_id = revision_id.encode('utf-8')
    inv = inventory.Inventory(root_id=None, revision_id=revision_id)
    for e in elt:
        ie = unpack_entry(e, entry_cache, return_from_cache)
        inv.add(ie)
    return inv


def serialize_inventory_flat(inv, append, root_id, supported_kinds, working):
    """Serialize an inventory to a flat XML file.

    :param inv: Inventory to serialize
    :param append: Function for writing a line of output
    :param working: If True skip history data - text_sha1, text_size,
        reference_revision, symlink_target.    self._check_revisions(inv)
    """
    entries = inv.iter_entries()
    # Skip the root
    root_path, root_ie = next(entries)
    for path, ie in entries:
        if ie.parent_id != root_id:
            parent_str = b''.join(
                [b' parent_id="', encode_and_escape(ie.parent_id), b'"'])
        else:
            parent_str = b''
        if ie.kind == 'file':
            if ie.executable:
                executable = b' executable="yes"'
            else:
                executable = b''
            if not working:
                append(b'<file%s file_id="%s" name="%s"%s revision="%s" '
                       b'text_sha1="%s" text_size="%d" />\n' % (
                           executable, encode_and_escape(ie.file_id),
                           encode_and_escape(ie.name), parent_str,
                           encode_and_escape(ie.revision), ie.text_sha1,
                           ie.text_size))
            else:
                append(b'<file%s file_id="%s" name="%s"%s />\n' % (
                    executable, encode_and_escape(ie.file_id),
                    encode_and_escape(ie.name), parent_str))
        elif ie.kind == 'directory':
            if not working:
                append(b'<directory file_id="%s" name="%s"%s revision="%s" '
                       b'/>\n' % (
                           encode_and_escape(ie.file_id),
                           encode_and_escape(ie.name),
                           parent_str,
                           encode_and_escape(ie.revision)))
            else:
                append(b'<directory file_id="%s" name="%s"%s />\n' % (
                    encode_and_escape(ie.file_id),
                    encode_and_escape(ie.name),
                    parent_str))
        elif ie.kind == 'symlink':
            if not working:
                append(b'<symlink file_id="%s" name="%s"%s revision="%s" '
                       b'symlink_target="%s" />\n' % (
                           encode_and_escape(ie.file_id),
                           encode_and_escape(ie.name),
                           parent_str,
                           encode_and_escape(ie.revision),
                           encode_and_escape(ie.symlink_target)))
            else:
                append(b'<symlink file_id="%s" name="%s"%s />\n' % (
                    encode_and_escape(ie.file_id),
                    encode_and_escape(ie.name),
                    parent_str))
        elif ie.kind == 'tree-reference':
            if ie.kind not in supported_kinds:
                raise serializer.UnsupportedInventoryKind(ie.kind)
            if not working:
                append(b'<tree-reference file_id="%s" name="%s"%s '
                       b'revision="%s" reference_revision="%s" />\n' % (
                           encode_and_escape(ie.file_id),
                           encode_and_escape(ie.name),
                           parent_str,
                           encode_and_escape(ie.revision),
                           encode_and_escape(ie.reference_revision)))
            else:
                append(b'<tree-reference file_id="%s" name="%s"%s />\n' % (
                    encode_and_escape(ie.file_id),
                    encode_and_escape(ie.name),
                    parent_str))
        else:
            raise serializer.UnsupportedInventoryKind(ie.kind)
    append(b'</inventory>\n')



# Copyright (C) 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Fixtures that can be used within tests.

Fixtures can be created during a test as a way to separate out creation of
objects to test.  Fixture objects can hold some state so that different
objects created during a test instance can be related.  Normally a fixture
should live only for the duration of a single test, and its tearDown method
should be passed to `addCleanup` on the test.
"""

import itertools
from contextlib import ExitStack


def generate_unicode_names():
    """Generate a sequence of arbitrary unique unicode names.

    By default they are not representable in ascii.

    >>> gen = generate_unicode_names()
    >>> n1 = next(gen)
    >>> n2 = next(gen)
    >>> type(n1)
    <class 'str'>
    >>> n1 == n2
    False
    >>> n1.encode('ascii', 'replace') == n1
    False
    """
    # include a mathematical symbol unlikely to be in 8-bit encodings
    return ("\N{SINE WAVE}%d" % x for x in itertools.count())


interesting_encodings = [
    ('iso-8859-1', False),
    ('ascii', False),
    ('cp850', False),
    ('utf-8', True),
    ('ucs-2', True),
    ]


def generate_unicode_encodings(universal_encoding=None):
    """Return a generator of unicode encoding names.

    These can be passed to Python encode/decode/etc.

    :param universal_encoding: True/False/None tristate to say whether the
        generated encodings either can or cannot encode all unicode
        strings.

    >>> n1 = next(generate_unicode_names())
    >>> enc = next(generate_unicode_encodings(universal_encoding=True))
    >>> enc2 = next(generate_unicode_encodings(universal_encoding=False))
    >>> n1.encode(enc).decode(enc) == n1
    True
    >>> try:
    ...   n1.encode(enc2).decode(enc2)
    ... except UnicodeError:
    ...   print('fail')
    fail
    """
    # TODO: check they're supported on this platform?
    if universal_encoding is not None:
        e = [n for (n, u) in interesting_encodings if u == universal_encoding]
    else:
        e = [n for (n, u) in interesting_encodings]
    return itertools.cycle(iter(e))


class RecordingContextManager:
    """A context manager that records."""

    def __init__(self):
        self._calls = []

    def __enter__(self):
        self._calls.append('__enter__')
        return self  # This is bound to the 'as' clause in a with statement.

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._calls.append('__exit__')
        return False  # propogate exceptions.


def build_branch_with_non_ancestral_rev(branch_builder):
    """Builds a branch with a rev not in the ancestry of the tip.

    This is the revision graph::

      rev-2
        |
      rev-1
        |
      (null)

    The branch tip is 'rev-1'.  'rev-2' is present in the branch's repository,
    but is not part of rev-1's ancestry.

    :param branch_builder: A BranchBuilder (e.g. from
        TestCaseWithMemoryTransport.make_branch_builder).
    :returns: the new branch
    """
    # Make a sequence of two commits
    rev1 = branch_builder.build_commit(message="Rev 1")
    rev2 = branch_builder.build_commit(message="Rev 2")
    # Move the branch tip back to the first commit
    source = branch_builder.get_branch()
    source.set_last_revision_info(1, rev1)
    return source, rev1, rev2


def make_branch_and_populated_tree(testcase):
    """Make a simple branch and tree.

    The tree holds some added but uncommitted files.
    """
    # TODO: Either accept or return the names of the files, so the caller
    # doesn't need to be bound to the particular files created? -- mbp
    # 20110705
    tree = testcase.make_branch_and_tree('t')
    testcase.build_tree_contents([('t/hello', b'hello world')])
    tree.add(['hello'], ids=[b'hello-id'])
    return tree


class TimeoutException(Exception):
    """Timeout expired"""


class TimeoutFixture:
    """Kill a test with sigalarm if it runs too long.

    Only works on Unix at present.
    """

    def __init__(self, timeout_secs, gentle=True):
        import signal
        self.timeout_secs = timeout_secs
        self.alarm_fn = getattr(signal, 'alarm', None)
        self.gentle = gentle
        self._es = ExitStack()

    def signal_handler(self, signum, frame):
        raise TimeoutException(self.timeout_secs)

    def setUp(self):
        import signal
        if self.alarm_fn is None:
            return
        if self.gentle:
            # Install a handler for SIGARLM so we can raise an exception rather
            # than the default handler executing, which kills the process.
            old_handler = signal.signal(signal.SIGALRM, self.signal_handler)
        # We add the alarm cleanup before the cleanup for the signal handler,
        # otherwise there is a race condition where the signal handler is
        # cleaned up but the alarm still fires.
        self._es.callback(self.alarm_fn, 0)
        self.alarm_fn(self.timeout_secs)
        if self.gentle:
            self._es.callback(signal.signal, signal.SIGALRM, old_handler)

    def cleanUp(self):
        self._es.close()



# Copyright (C) 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Inventory/revision serialization."""

from .. import errors, registry


class BadInventoryFormat(errors.BzrError):

    _fmt = "Root class for inventory serialization errors"


class UnexpectedInventoryFormat(BadInventoryFormat):

    _fmt = "The inventory was not in the expected format:\n %(msg)s"

    def __init__(self, msg):
        BadInventoryFormat.__init__(self, msg=msg)


class UnsupportedInventoryKind(errors.BzrError):

    _fmt = """Unsupported entry kind %(kind)s"""

    def __init__(self, kind):
        self.kind = kind


class Serializer:
    """Inventory and revision serialization/deserialization."""

    squashes_xml_invalid_characters = False

    def write_inventory(self, inv, f):
        """Write inventory to a file.

        Note: this is a *whole inventory* operation, and should only be used
        sparingly, as it does not scale well with large trees.
        """
        raise NotImplementedError(self.write_inventory)

    def write_inventory_to_chunks(self, inv):
        """Produce a simple bytestring chunk representation of an inventory.

        Note: this is a *whole inventory* operation, and should only be used
        sparingly, as it does not scale well with large trees.

        The requirement for the contents of the string is that it can be passed
        to read_inventory_from_lines and the result is an identical inventory
        in memory.
        """
        raise NotImplementedError(self.write_inventory_to_chunks)

    def write_inventory_to_lines(self, inv):
        """Produce a simple lines representation of an inventory.

        Note: this is a *whole inventory* operation, and should only be used
        sparingly, as it does not scale well with large trees.

        The requirement for the contents of the string is that it can be passed
        to read_inventory_from_lines and the result is an identical inventory
        in memory.
        """
        raise NotImplementedError(self.write_inventory_to_lines)

    def read_inventory_from_lines(self, lines, revision_id=None,
                                  entry_cache=None, return_from_cache=False):
        """Read bytestring chunks into an inventory object.

        :param lines: The serialized inventory to read.
        :param revision_id: If not-None, the expected revision id of the
            inventory. Some serialisers use this to set the results' root
            revision. This should be supplied for deserialising all
            from-repository inventories so that xml5 inventories that were
            serialised without a revision identifier can be given the right
            revision id (but not for working tree inventories where users can
            edit the data without triggering checksum errors or anything).
        :param entry_cache: An optional cache of InventoryEntry objects. If
            supplied we will look up entries via (file_id, revision_id) which
            should map to a valid InventoryEntry (File/Directory/etc) object.
        :param return_from_cache: Return entries directly from the cache,
            rather than copying them first. This is only safe if the caller
            promises not to mutate the returned inventory entries, but it can
            make some operations significantly faster.
        """
        raise NotImplementedError(self.read_inventory_from_lines)

    def read_inventory(self, f, revision_id=None):
        """See read_inventory_from_lines."""
        raise NotImplementedError(self.read_inventory)

    def write_revision_to_string(self, rev):
        raise NotImplementedError(self.write_revision_to_string)

    def write_revision_to_lines(self, rev):
        raise NotImplementedError(self.write_revision_to_lines)

    def read_revision(self, f):
        raise NotImplementedError(self.read_revision)

    def read_revision_from_string(self, xml_string):
        raise NotImplementedError(self.read_revision_from_string)


class SerializerRegistry(registry.Registry):
    """Registry for serializer objects"""


format_registry = SerializerRegistry()
format_registry.register_lazy('5', 'breezy.bzr.xml5', 'serializer_v5')
format_registry.register_lazy('6', 'breezy.bzr.xml6', 'serializer_v6')
format_registry.register_lazy('7', 'breezy.bzr.xml7', 'serializer_v7')
format_registry.register_lazy('8', 'breezy.bzr.xml8', 'serializer_v8')
format_registry.register_lazy('9', 'breezy.bzr.chk_serializer',
                              'chk_serializer_255_bigpage')
format_registry.register_lazy('10', 'breezy.bzr.chk_serializer',
                              'chk_bencode_serializer')



# Copyright (C) 2005-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""File annotate based on weave storage"""

# TODO: Choice of more or less verbose formats:
#
# interposed: show more details between blocks of modified lines

# TODO: Show which revision caused a line to merge into the parent

# TODO: perhaps abbreviate timescales depending on how recent they are
# e.g. "3:12 Tue", "13 Oct", "Oct 2005", etc.

import sys
import time

from .lazy_import import lazy_import

lazy_import(globals(), """
import patiencediff

from breezy import (
    tsort,
    )
""")
from . import config, errors, osutils
from .repository import _strip_NULL_ghosts
from .revision import CURRENT_REVISION, Revision


def annotate_file_tree(tree, path, to_file, verbose=False, full=False,
                       show_ids=False, branch=None):
    """Annotate path in a tree.

    The tree should already be read_locked() when annotate_file_tree is called.

    :param tree: The tree to look for revision numbers and history from.
    :param path: The path to annotate
    :param to_file: The file to output the annotation to.
    :param verbose: Show all details rather than truncating to ensure
        reasonable text width.
    :param full: XXXX Not sure what this does.
    :param show_ids: Show revision ids in the annotation output.
    :param branch: Branch to use for revision revno lookups
    """
    if branch is None:
        branch = tree.branch
    if to_file is None:
        to_file = sys.stdout

    encoding = osutils.get_terminal_encoding()
    # Handle the show_ids case
    annotations = list(tree.annotate_iter(path))
    if show_ids:
        return _show_id_annotations(annotations, to_file, full, encoding)

    if not getattr(tree, "get_revision_id", False):
        # Create a virtual revision to represent the current tree state.
        # Should get some more pending commit attributes, like pending tags,
        # bugfixes etc.
        current_rev = Revision(CURRENT_REVISION)
        current_rev.parent_ids = tree.get_parent_ids()
        try:
            current_rev.committer = branch.get_config_stack().get('email')
        except errors.NoWhoami:
            current_rev.committer = 'local user'
        current_rev.message = "?"
        current_rev.timestamp = round(time.time(), 3)
        current_rev.timezone = osutils.local_time_offset()
    else:
        current_rev = None
    annotation = list(_expand_annotations(
        annotations, branch, current_rev))
    _print_annotations(annotation, verbose, to_file, full, encoding)


def _print_annotations(annotation, verbose, to_file, full, encoding):
    """Print annotations to to_file.

    :param to_file: The file to output the annotation to.
    :param verbose: Show all details rather than truncating to ensure
        reasonable text width.
    :param full: XXXX Not sure what this does.
    """
    if len(annotation) == 0:
        max_origin_len = max_revno_len = 0
    else:
        max_origin_len = max(len(x[1]) for x in annotation)
        max_revno_len = max(len(x[0]) for x in annotation)
    if not verbose:
        max_revno_len = min(max_revno_len, 12)
    max_revno_len = max(max_revno_len, 3)

    # Output the annotations
    prevanno = ''
    for (revno_str, author, date_str, line_rev_id, text) in annotation:
        if verbose:
            anno = '%-*s %-*s %8s ' % (max_revno_len, revno_str,
                                       max_origin_len, author, date_str)
        else:
            if len(revno_str) > max_revno_len:
                revno_str = revno_str[:max_revno_len - 1] + '>'
            anno = "%-*s %-7s " % (max_revno_len, revno_str, author[:7])
        if anno.lstrip() == "" and full:
            anno = prevanno
        # GZ 2017-05-21: Writing both unicode annotation and bytes from file
        # which the given to_file must cope with.
        to_file.write(anno)
        to_file.write('| {}\n'.format(text.decode(encoding)))
        prevanno = anno


def _show_id_annotations(annotations, to_file, full, encoding):
    if not annotations:
        return
    last_rev_id = None
    max_origin_len = max(len(origin) for origin, text in annotations)
    for origin, text in annotations:
        if full or last_rev_id != origin:
            this = origin
        else:
            this = b''
        to_file.write('%*s | %s' % (
            max_origin_len, this.decode('utf-8'), text.decode(encoding)))
        last_rev_id = origin
    return


def _expand_annotations(annotations, branch, current_rev=None):
    """Expand a file's annotations into command line UI ready tuples.

    Each tuple includes detailed information, such as the author name, and date
    string for the commit, rather than just the revision id.

    :param annotations: The annotations to expand.
    :param revision_id_to_revno: A map from id to revision numbers.
    :param branch: A locked branch to query for revision details.
    """
    repository = branch.repository
    revision_ids = {o for o, t in annotations}
    if current_rev is not None:
        # This can probably become a function on MutableTree, get_revno_map
        # there, or something.
        last_revision = current_rev.revision_id
        # XXX: Partially Cloned from branch, uses the old_get_graph, eep.
        # XXX: The main difficulty is that we need to inject a single new node
        #      (current_rev) into the graph before it gets numbered, etc.
        #      Once KnownGraph gets an 'add_node()' function, we can use
        #      VF.get_known_graph_ancestry().
        graph = repository.get_graph()
        revision_graph = {
            key: value for key, value in
            graph.iter_ancestry(current_rev.parent_ids) if value is not None}
        revision_graph = _strip_NULL_ghosts(revision_graph)
        revision_graph[last_revision] = current_rev.parent_ids
        merge_sorted_revisions = tsort.merge_sort(
            revision_graph,
            last_revision,
            None,
            generate_revno=True)
        revision_id_to_revno = {
            rev_id: revno
            for seq_num, rev_id, depth, revno, end_of_merge in
            merge_sorted_revisions}
    else:
        # TODO(jelmer): Only look up the revision ids that we need (i.e. those
        # in revision_ids). Possibly add a HPSS call that can look those up
        # in bulk over HPSS.
        revision_id_to_revno = branch.get_revision_id_to_revno_map()
    last_origin = None
    revisions = {}
    if CURRENT_REVISION in revision_ids:
        revision_id_to_revno[CURRENT_REVISION] = (
            "%d?" % (branch.revno() + 1),)
        revisions[CURRENT_REVISION] = current_rev
    revisions.update(
        entry for entry in
        repository.iter_revisions(revision_ids)
        if entry[1] is not None)
    for origin, text in annotations:
        text = text.rstrip(b'\r\n')
        if origin == last_origin:
            (revno_str, author, date_str) = ('', '', '')
        else:
            last_origin = origin
            if origin not in revisions:
                (revno_str, author, date_str) = ('?', '?', '?')
            else:
                revno_str = '.'.join(
                    str(i) for i in revision_id_to_revno[origin])
            rev = revisions[origin]
            tz = rev.timezone or 0
            date_str = time.strftime('%Y%m%d',
                                     time.gmtime(rev.timestamp + tz))
            # a lazy way to get something like the email address
            # TODO: Get real email address
            author = rev.get_apparent_authors()[0]
            _, email = config.parse_username(author)
            if email:
                author = email
        yield (revno_str, author, date_str, origin, text)


def reannotate(parents_lines, new_lines, new_revision_id,
               _left_matching_blocks=None,
               heads_provider=None):
    """Create a new annotated version from new lines and parent annotations.

    :param parents_lines: List of annotated lines for all parents
    :param new_lines: The un-annotated new lines
    :param new_revision_id: The revision-id to associate with new lines
        (will often be CURRENT_REVISION)
    :param left_matching_blocks: a hint about which areas are common
        between the text and its left-hand-parent.  The format is
        the SequenceMatcher.get_matching_blocks format
        (start_left, start_right, length_of_match).
    :param heads_provider: An object which provides a .heads() call to resolve
        if any revision ids are children of others.
        If None, then any ancestry disputes will be resolved with
        new_revision_id
    """
    if len(parents_lines) == 0:
        lines = [(new_revision_id, line) for line in new_lines]
    elif len(parents_lines) == 1:
        lines = _reannotate(parents_lines[0], new_lines, new_revision_id,
                            _left_matching_blocks)
    elif len(parents_lines) == 2:
        left = _reannotate(parents_lines[0], new_lines, new_revision_id,
                           _left_matching_blocks)
        lines = _reannotate_annotated(parents_lines[1], new_lines,
                                      new_revision_id, left,
                                      heads_provider)
    else:
        reannotations = [_reannotate(parents_lines[0], new_lines,
                                     new_revision_id, _left_matching_blocks)]
        reannotations.extend(_reannotate(p, new_lines, new_revision_id)
                             for p in parents_lines[1:])
        lines = []
        for annos in zip(*reannotations):
            origins = {a for a, l in annos}
            if len(origins) == 1:
                # All the parents agree, so just return the first one
                lines.append(annos[0])
            else:
                line = annos[0][1]
                if len(origins) == 2 and new_revision_id in origins:
                    origins.remove(new_revision_id)
                if len(origins) == 1:
                    lines.append((origins.pop(), line))
                else:
                    lines.append((new_revision_id, line))
    return lines


def _reannotate(parent_lines, new_lines, new_revision_id,
                matching_blocks=None):
    new_cur = 0
    if matching_blocks is None:
        plain_parent_lines = [l for r, l in parent_lines]
        matcher = patiencediff.PatienceSequenceMatcher(
            None, plain_parent_lines, new_lines)
        matching_blocks = matcher.get_matching_blocks()
    lines = []
    for i, j, n in matching_blocks:
        for line in new_lines[new_cur:j]:
            lines.append((new_revision_id, line))
        lines.extend(parent_lines[i:i + n])
        new_cur = j + n
    return lines


def _get_matching_blocks(old, new):
    matcher = patiencediff.PatienceSequenceMatcher(None, old, new)
    return matcher.get_matching_blocks()


_break_annotation_tie = None


def _old_break_annotation_tie(annotated_lines):
    """Chose an attribution between several possible ones.

    :param annotated_lines: A list of tuples ((file_id, rev_id), line) where
        the lines are identical but the revids different while no parent
        relation exist between them

     :return : The "winning" line. This must be one with a revid that
         guarantees that further criss-cross merges will converge. Failing to
         do so have performance implications.
    """
    # sort lexicographically so that we always get a stable result.

    # TODO: while 'sort' is the easiest (and nearly the only possible solution)
    # with the current implementation, chosing the oldest revision is known to
    # provide better results (as in matching user expectations). The most
    # common use case being manual cherry-pick from an already existing
    # revision.
    return sorted(annotated_lines)[0]


def _find_matching_unannotated_lines(output_lines, plain_child_lines,
                                     child_lines, start_child, end_child,
                                     right_lines, start_right, end_right,
                                     heads_provider, revision_id):
    """Find lines in plain_right_lines that match the existing lines.

    :param output_lines: Append final annotated lines to this list
    :param plain_child_lines: The unannotated new lines for the child text
    :param child_lines: Lines for the child text which have been annotated
        for the left parent

    :param start_child: Position in plain_child_lines and child_lines to start
        the match searching
    :param end_child: Last position in plain_child_lines and child_lines to
        search for a match
    :param right_lines: The annotated lines for the whole text for the right
        parent
    :param start_right: Position in right_lines to start the match
    :param end_right: Last position in right_lines to search for a match
    :param heads_provider: When parents disagree on the lineage of a line, we
        need to check if one side supersedes the other
    :param revision_id: The label to give if a line should be labeled 'tip'
    """
    output_extend = output_lines.extend
    output_append = output_lines.append
    # We need to see if any of the unannotated lines match
    plain_right_subset = [l for a, l in right_lines[start_right:end_right]]
    plain_child_subset = plain_child_lines[start_child:end_child]
    match_blocks = _get_matching_blocks(plain_right_subset, plain_child_subset)

    last_child_idx = 0

    for right_idx, child_idx, match_len in match_blocks:
        # All the lines that don't match are just passed along
        if child_idx > last_child_idx:
            output_extend(child_lines[start_child + last_child_idx:
                                      start_child + child_idx])
        for offset in range(match_len):
            left = child_lines[start_child + child_idx + offset]
            right = right_lines[start_right + right_idx + offset]
            if left[0] == right[0]:
                # The annotations match, just return the left one
                output_append(left)
            elif left[0] == revision_id:
                # The left parent marked this as unmatched, so let the
                # right parent claim it
                output_append(right)
            else:
                # Left and Right both claim this line
                if heads_provider is None:
                    output_append((revision_id, left[1]))
                else:
                    heads = heads_provider.heads((left[0], right[0]))
                    if len(heads) == 1:
                        output_append((next(iter(heads)), left[1]))
                    else:
                        # Both claim different origins, get a stable result.
                        # If the result is not stable, there is a risk a
                        # performance degradation as criss-cross merges will
                        # flip-flop the attribution.
                        if _break_annotation_tie is None:
                            output_append(
                                _old_break_annotation_tie([left, right]))
                        else:
                            output_append(_break_annotation_tie([left, right]))
        last_child_idx = child_idx + match_len


def _reannotate_annotated(right_parent_lines, new_lines, new_revision_id,
                          annotated_lines, heads_provider):
    """Update the annotations for a node based on another parent.

    :param right_parent_lines: A list of annotated lines for the right-hand
        parent.
    :param new_lines: The unannotated new lines.
    :param new_revision_id: The revision_id to attribute to lines which are not
        present in either parent.
    :param annotated_lines: A list of annotated lines. This should be the
        annotation of new_lines based on parents seen so far.
    :param heads_provider: When parents disagree on the lineage of a line, we
        need to check if one side supersedes the other.
    """
    if len(new_lines) != len(annotated_lines):
        raise AssertionError("mismatched new_lines and annotated_lines")
    # First compare the newly annotated lines with the right annotated lines.
    # Lines which were not changed in left or right should match. This tends to
    # be the bulk of the lines, and they will need no further processing.
    lines = []
    lines_extend = lines.extend
    # The line just after the last match from the right side
    last_right_idx = 0
    last_left_idx = 0
    matching_left_and_right = _get_matching_blocks(right_parent_lines,
                                                   annotated_lines)
    for right_idx, left_idx, match_len in matching_left_and_right:
        # annotated lines from last_left_idx to left_idx did not match the
        # lines from last_right_idx to right_idx, the raw lines should be
        # compared to determine what annotations need to be updated
        if last_right_idx == right_idx or last_left_idx == left_idx:
            # One of the sides is empty, so this is a pure insertion
            lines_extend(annotated_lines[last_left_idx:left_idx])
        else:
            # We need to see if any of the unannotated lines match
            _find_matching_unannotated_lines(lines,
                                             new_lines, annotated_lines,
                                             last_left_idx, left_idx,
                                             right_parent_lines,
                                             last_right_idx, right_idx,
                                             heads_provider,
                                             new_revision_id)
        last_right_idx = right_idx + match_len
        last_left_idx = left_idx + match_len
        # If left and right agree on a range, just push that into the output
        lines_extend(annotated_lines[left_idx:left_idx + match_len])
    return lines


try:
    from breezy._annotator_pyx import Annotator
except ImportError as e:
    osutils.failed_to_load_extension(e)
    from breezy._annotator_py import Annotator  # noqa: F401



# Copyright (C) 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


import os

from breezy import tests
from breezy.tests import features


class TestTreeShape(tests.TestCaseWithTransport):

    def test_build_tree(self):
        """Test tree-building test helper"""
        self.build_tree_contents([
            ('foo', b'new contents'),
            ('.bzr/',),
            ('.bzr/README', b'hello'),
            ])
        self.assertPathExists('foo')
        self.assertPathExists('.bzr/README')
        self.assertFileEqual(b'hello', '.bzr/README')

    def test_build_tree_symlink(self):
        self.requireFeature(features.SymlinkFeature(self.test_dir))
        self.build_tree_contents([('link@', 'target')])
        self.assertEqual('target',
                         os.readlink('link'))



# Copyright (C) 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Fastimport/fastexport commands."""

from ... import controldir
from ...commands import Command
from ...option import Option, RegistryOption
from . import helpers, load_fastimport


def _run(source, processor_factory, verbose=False, user_map=None, **kwargs):
    """Create and run a processor.

    :param source: a filename or '-' for standard input. If the
      filename ends in .gz, it will be opened as a gzip file and
      the stream will be implicitly uncompressed
    :param processor_factory: a callable for creating a processor
    :param user_map: if not None, the file containing the user map.
    """
    from fastimport import parser
    from fastimport.errors import ParsingError

    from ...errors import CommandError
    stream = _get_source_stream(source)
    user_mapper = _get_user_mapper(user_map)
    proc = processor_factory(verbose=verbose, **kwargs)
    p = parser.ImportParser(stream, verbose=verbose, user_mapper=user_mapper)
    try:
        return proc.process(p.iter_commands)
    except ParsingError as e:
        raise CommandError("%d: Parse error: %s" % (e.lineno, e))


def _get_source_stream(source):
    if source == '-' or source is None:
        import sys
        try:
            stream = sys.stdin.buffer
        except AttributeError:
            stream = helpers.binary_stream(sys.stdin)
    elif source.endswith('.gz'):
        import gzip
        stream = gzip.open(source, "rb")
    else:
        stream = open(source, "rb")
    return stream


def _get_user_mapper(filename):
    from . import user_mapper
    if filename is None:
        return None
    f = open(filename)
    lines = f.readlines()
    f.close()
    return user_mapper.UserMapper(lines)


class cmd_fast_import(Command):
    """Backend for fast Bazaar data importers.

    This command reads a mixed command/data stream and creates
    branches in a Bazaar repository accordingly. The preferred
    recipe is::

      bzr fast-import project.fi project.bzr

    Numerous commands are provided for generating a fast-import file
    to use as input.
    To specify standard input as the input stream, use a
    source name of '-' (instead of project.fi). If the source name
    ends in '.gz', it is assumed to be compressed in gzip format.

    project.bzr will be created if it doesn't exist. If it exists
    already, it should be empty or be an existing Bazaar repository
    or branch. If not specified, the current directory is assumed.

    fast-import will intelligently select the format to use when
    creating a repository or branch. If you are running Bazaar 1.17
    up to Bazaar 2.0, the default format for Bazaar 2.x ("2a") is used.
    Otherwise, the current default format ("pack-0.92" for Bazaar 1.x)
    is used. If you wish to specify a custom format, use the `--format`
    option.

     .. note::

        To maintain backwards compatibility, fast-import lets you
        create the target repository or standalone branch yourself.
        It is recommended though that you let fast-import create
        these for you instead.

    :Branch mapping rules:

     Git reference names are mapped to Bazaar branch names as follows:

     * refs/heads/foo is mapped to foo
     * refs/remotes/origin/foo is mapped to foo.remote
     * refs/tags/foo is mapped to foo.tag
     * */master is mapped to trunk, trunk.remote, etc.
     * */trunk is mapped to git-trunk, git-trunk.remote, etc.

    :Branch creation rules:

     When a shared repository is created or found at the destination,
     branches are created inside it. In the simple case of a single
     branch (refs/heads/master) inside the input file, the branch is
     project.bzr/trunk.

     When a standalone branch is found at the destination, the trunk
     is imported there and warnings are output about any other branches
     found in the input file.

     When a branch in a shared repository is found at the destination,
     that branch is made the trunk and other branches, if any, are
     created in sister directories.

    :Working tree updates:

     The working tree is generated for the trunk branch. If multiple
     branches are created, a message is output on completion explaining
     how to create the working trees for other branches.

    :Custom exporters:

     The fast-export-from-xxx commands typically call more advanced
     xxx-fast-export scripts. You are welcome to use the advanced
     scripts if you prefer.

     If you wish to write a custom exporter for your project, see
     http://bazaar-vcs.org/BzrFastImport for the detailed protocol
     specification. In many cases, exporters can be written quite
     quickly using whatever scripting/programming language you like.

    :User mapping:

     Some source repositories store just the user name while Bazaar
     prefers a full email address. You can adjust user-ids while
     importing by using the --user-map option. The argument is a
     text file with lines in the format::

       old-id = new-id

     Blank lines and lines beginning with # are ignored.
     If old-id has the special value '@', then users without an
     email address will get one created by using the matching new-id
     as the domain, unless a more explicit address is given for them.
     For example, given the user-map of::

       @ = example.com
       bill = William Jones <bill@example.com>

     then user-ids are mapped as follows::

      maria => maria <maria@example.com>
      bill => William Jones <bill@example.com>

     .. note::

        User mapping is supported by both the fast-import and
        fast-import-filter commands.

    :Blob tracking:

     As some exporters (like git-fast-export) reuse blob data across
     commits, fast-import makes two passes over the input file by
     default. In the first pass, it collects data about what blobs are
     used when, along with some other statistics (e.g. total number of
     commits). In the second pass, it generates the repository and
     branches.

     .. note::

        The initial pass isn't done if the --info option is used
        to explicitly pass in information about the input stream.
        It also isn't done if the source is standard input. In the
        latter case, memory consumption may be higher than otherwise
        because some blobs may be kept in memory longer than necessary.

    :Restarting an import:

     At checkpoints and on completion, the commit-id -> revision-id
     map is saved to a file called 'fastimport-id-map' in the control
     directory for the repository (e.g. .bzr/repository). If the import
     is interrupted or unexpectedly crashes, it can be started again
     and this file will be used to skip over already loaded revisions.
     As long as subsequent exports from the original source begin
     with exactly the same revisions, you can use this feature to
     maintain a mirror of a repository managed by a foreign tool.
     If and when Bazaar is used to manage the repository, this file
     can be safely deleted.

    :Examples:

     Import a Subversion repository into Bazaar::

       svn-fast-export /svn/repo/path > project.fi
       bzr fast-import project.fi project.bzr

     Import a CVS repository into Bazaar::

       cvs2git /cvs/repo/path > project.fi
       bzr fast-import project.fi project.bzr

     Import a Git repository into Bazaar::

       cd /git/repo/path
       git fast-export --all > project.fi
       bzr fast-import project.fi project.bzr

     Import a Mercurial repository into Bazaar::

       cd /hg/repo/path
       hg fast-export > project.fi
       bzr fast-import project.fi project.bzr

     Import a Darcs repository into Bazaar::

       cd /darcs/repo/path
       darcs-fast-export > project.fi
       bzr fast-import project.fi project.bzr
    """
    hidden = False
    _see_also = ['fast-export', 'fast-import-filter', 'fast-import-info']
    takes_args = ['source', 'destination?']
    takes_options = ['verbose',
                     Option('user-map', type=str,
                            help="Path to file containing a map of user-ids.",
                            ),
                     Option('info', type=str,
                            help="Path to file containing caching hints.",
                            ),
                     Option('trees',
                            help="Update all working trees, not just trunk's.",
                            ),
                     Option('count', type=int,
                            help="Import this many revisions then exit.",
                            ),
                     Option('checkpoint', type=int,
                            help="Checkpoint automatically every N revisions."
                            " The default is 10000.",
                            ),
                     Option('autopack', type=int,
                            help="Pack every N checkpoints. The default is 4.",
                            ),
                     Option('inv-cache', type=int,
                            help="Number of inventories to cache.",
                            ),
                     RegistryOption.from_kwargs('mode',
                                                'The import algorithm to use.',
                                                title='Import Algorithm',
                                                default='Use the preferred algorithm (inventory deltas).',
                                                experimental="Enable experimental features.",
                                                value_switches=True, enum_switch=False,
                                                ),
                     Option('import-marks', type=str,
                            help="Import marks from file."
                            ),
                     Option('export-marks', type=str,
                            help="Export marks to file."
                            ),
                     RegistryOption('format',
                                    help='Specify a format for the created repository. See'
                                    ' "bzr help formats" for details.',
                                    lazy_registry=(
                                        'breezy.controldir', 'format_registry'),
                                    converter=lambda name: controldir.format_registry.make_controldir(
                                        name),
                                    value_switches=False, title='Repository format'),
                     ]

    def run(self, source, destination='.', verbose=False, info=None,
            trees=False, count=-1, checkpoint=10000, autopack=4, inv_cache=-1,
            mode=None, import_marks=None, export_marks=None, format=None,
            user_map=None):
        load_fastimport()
        from .helpers import open_destination_directory
        from .processors import generic_processor
        control = open_destination_directory(destination, format=format)

        # If an information file was given and the source isn't stdin,
        # generate the information by reading the source file as a first pass
        if info is None and source != '-':
            info = self._generate_info(source)

        # Do the work
        if mode is None:
            mode = 'default'
        params = {
            'info': info,
            'trees': trees,
            'count': count,
            'checkpoint': checkpoint,
            'autopack': autopack,
            'inv-cache': inv_cache,
            'mode': mode,
            'import-marks': import_marks,
            'export-marks': export_marks,
            }
        return _run(source, generic_processor.GenericProcessor,
                    bzrdir=control, params=params, verbose=verbose,
                    user_map=user_map)

    def _generate_info(self, source):
        from io import StringIO

        from fastimport import parser
        from fastimport.errors import ParsingError
        from fastimport.processors import info_processor

        from ...errors import CommandError
        stream = _get_source_stream(source)
        output = StringIO()
        try:
            proc = info_processor.InfoProcessor(verbose=True, outf=output)
            p = parser.ImportParser(stream)
            try:
                return_code = proc.process(p.iter_commands)
            except ParsingError as e:
                raise CommandError("%d: Parse error: %s" % (e.lineno, e))
            lines = output.getvalue().splitlines()
        finally:
            output.close()
            stream.seek(0)
        return lines


class cmd_fast_export(Command):
    """Generate a fast-import stream from a Bazaar branch.

    This program generates a stream from a Bazaar branch in fast-import
    format used by tools such as bzr fast-import, git-fast-import and
    hg-fast-import.

    It takes two optional arguments: the source bzr branch to export and
    the destination to write the file to write the fastimport stream to.

    If no source is specified, it will search for a branch in the
    current directory.

    If no destination is given or the destination is '-', standard output
    is used. Otherwise, the destination is the name of a file. If the
    destination ends in '.gz', the output will be compressed into gzip
    format.

    :Round-tripping:

     Recent versions of the fast-import specification support features
     that allow effective round-tripping most of the metadata in Bazaar
     branches. As such, fast-exporting a branch and fast-importing the data
     produced will create a new repository with roughly equivalent history, i.e.
     "bzr log -v -p --include-merges --forward" on the old branch and
     new branch should produce similar, if not identical, results.

     .. note::

        Be aware that the new repository may appear to have similar history
        but internally it is quite different with new revision-ids and
        file-ids assigned. As a consequence, the ability to easily merge
        with branches based on the old repository is lost. Depending on your
        reasons for producing a new repository, this may or may not be an
        issue.

    :Interoperability:

     fast-export can use the following "extended features" to
     produce a richer data stream:

     * *multiple-authors* - if a commit has multiple authors (as commonly
       occurs in pair-programming), all authors will be included in the
       output, not just the first author

     * *commit-properties* - custom metadata per commit that Bazaar stores
       in revision properties (e.g. branch-nick and bugs fixed by this
       change) will be included in the output.

     * *empty-directories* - directories, even the empty ones, will be
       included in the output.

     To disable these features and produce output acceptable to git 1.6,
     use the --plain option. To enable these features, use --no-plain.
     Currently, --plain is the default but that will change in the near
     future once the feature names and definitions are formally agreed
     to by the broader fast-import developer community.

     Git has stricter naming rules for tags and fast-export --plain
     will skip tags which can't be imported into git. To replace characters
     unsupported in git with an underscore instead, specify
     --rewrite-tag-names.

    :History truncation:

     It is sometimes convenient to simply truncate the revision history at a
     certain point.  The --baseline option, to be used in conjunction with -r,
     emits a baseline commit containing the state of the entire source tree at
     the first requested revision.  This allows a user to produce a tree
     identical to the original without munging multiple exports.

    :Examples:

     To produce data destined for import into Bazaar::

       bzr fast-export --no-plain my-bzr-branch my.fi.gz

     To produce data destined for Git 1.6::

       bzr fast-export --plain my-bzr-branch my.fi

     To import several unmerged but related branches into the same repository,
     use the --{export,import}-marks options, and specify a name for the git
     branch like this::

       bzr fast-export --export-marks=marks.bzr project.dev |
              GIT_DIR=project/.git git-fast-import --export-marks=marks.git

       bzr fast-export --import-marks=marks.bzr -b other project.other |
              GIT_DIR=project/.git git-fast-import --import-marks=marks.git

     If you get a "Missing space after source" error from git-fast-import,
     see the top of the commands.py module for a work-around.

     Since bzr uses per-branch tags and git/hg use per-repo tags, the
     way bzr fast-export presently emits tags (unconditional reset &
     new ref) may result in clashes when several different branches
     are imported into single git/hg repo.  If this occurs, use the
     bzr fast-export option --no-tags during the export of one or more
     branches to avoid the issue.
    """
    hidden = False
    _see_also = ['fast-import', 'fast-import-filter']
    takes_args = ['source?', 'destination?']
    takes_options = ['verbose', 'revision',
                     Option('git-branch', short_name='b', type=str,
                            argname='FILE',
                            help='Name of the git branch to create (default=master).'
                            ),
                     Option('checkpoint', type=int, argname='N',
                            help="Checkpoint every N revisions (default=10000)."
                            ),
                     Option('marks', type=str, argname='FILE',
                            help="Import marks from and export marks to file."
                            ),
                     Option('import-marks', type=str, argname='FILE',
                            help="Import marks from file."
                            ),
                     Option('export-marks', type=str, argname='FILE',
                            help="Export marks to file."
                            ),
                     Option('plain',
                            help="Exclude metadata to maximise interoperability."
                            ),
                     Option('rewrite-tag-names',
                            help="Replace characters invalid in git with '_'"
                            " (plain mode only).",
                            ),
                     Option('baseline',
                            help="Export an 'absolute' baseline commit prior to"
                            "the first relative commit",
                            ),
                     Option('no-tags',
                            help="Don't export tags"
                            ),
                     ]
    encoding_type = 'exact'

    def run(self, source=None, destination=None, verbose=False,
            git_branch="master", checkpoint=10000, marks=None,
            import_marks=None, export_marks=None, revision=None,
            plain=True, rewrite_tag_names=False, no_tags=False, baseline=False):
        load_fastimport()
        from ...branch import Branch
        from . import exporter

        if marks:
            import_marks = export_marks = marks

        # Open the source
        if source is None:
            source = "."
        branch = Branch.open_containing(source)[0]
        outf = exporter._get_output_stream(destination)
        exporter = exporter.BzrFastExporter(
            branch,
            outf=outf, ref=b"refs/heads/%s" % git_branch.encode('utf-8'),
            checkpoint=checkpoint, import_marks_file=import_marks,
            export_marks_file=export_marks, revision=revision, verbose=verbose,
            plain_format=plain, rewrite_tags=rewrite_tag_names,
            no_tags=no_tags, baseline=baseline)
        return exporter.run()



# Copyright (C) 2007-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

__docformat__ = "google"

"""Support for plugin hooking logic."""

from typing import Dict, List, Tuple

from . import errors, registry
from .lazy_import import lazy_import

lazy_import(globals(), """
from breezy import (
    _format_version_tuple,
    pyutils,
    )
from breezy.i18n import gettext
""")


class UnknownHook(errors.BzrError):

    _fmt = "The %(type)s hook '%(hook)s' is unknown in this version of breezy."

    def __init__(self, hook_type, hook_name):
        errors.BzrError.__init__(self)
        self.type = hook_type
        self.hook = hook_name


class KnownHooksRegistry(registry.Registry[str, "Hooks"]):
    # known_hooks registry contains
    # tuple of (module, member name) which is the hook point
    # module where the specific hooks are defined
    # callable to get the empty specific Hooks for that attribute

    def register_lazy_hook(self, hook_module_name, hook_member_name,
                           hook_factory_member_name):
        self.register_lazy((hook_module_name, hook_member_name),
                           hook_module_name, hook_factory_member_name)

    def iter_parent_objects(self):
        """Yield (hook_key, (parent_object, attr)) tuples for every registered
        hook, where 'parent_object' is the object that holds the hook
        instance.

        This is useful for resetting/restoring all the hooks to a known state,
        as is done in breezy.tests.TestCase._clear_hooks.
        """
        for key in self.keys():
            yield key, self.key_to_parent_and_attribute(key)

    def key_to_parent_and_attribute(self, key):
        """Convert a known_hooks key to a (parent_obj, attr) pair.

        :param key: A tuple (module_name, member_name) as found in the keys of
            the known_hooks registry.
        :return: The parent_object of the hook and the name of the attribute on
            that parent object where the hook is kept.
        """
        parent_mod, parent_member, attr = pyutils.calc_parent_name(*key)
        return pyutils.get_named_object(parent_mod, parent_member), attr


_builtin_known_hooks = (
    ('breezy.branch', 'Branch.hooks', 'BranchHooks'),
    ('breezy.controldir', 'ControlDir.hooks', 'ControlDirHooks'),
    ('breezy.commands', 'Command.hooks', 'CommandHooks'),
    ('breezy.config', 'ConfigHooks', '_ConfigHooks'),
    ('breezy.info', 'hooks', 'InfoHooks'),
    ('breezy.lock', 'Lock.hooks', 'LockHooks'),
    ('breezy.merge', 'Merger.hooks', 'MergeHooks'),
    ('breezy.msgeditor', 'hooks', 'MessageEditorHooks'),
    ('breezy.mutabletree', 'MutableTree.hooks', 'MutableTreeHooks'),
    ('breezy.bzr.smart.client', '_SmartClient.hooks', 'SmartClientHooks'),
    ('breezy.bzr.smart.server', 'SmartTCPServer.hooks', 'SmartServerHooks'),
    ('breezy.status', 'hooks', 'StatusHooks'),
    ('breezy.transport', 'Transport.hooks', 'TransportHooks'),
    ('breezy.version_info_formats.format_rio', 'RioVersionInfoBuilder.hooks',
        'RioVersionInfoBuilderHooks'),
    ('breezy.merge_directive', 'BaseMergeDirective.hooks',
        'MergeDirectiveHooks'),
    )

known_hooks = KnownHooksRegistry()
for (_hook_module, _hook_attribute, _hook_class) in _builtin_known_hooks:
    known_hooks.register_lazy_hook(_hook_module, _hook_attribute, _hook_class)
del _builtin_known_hooks, _hook_module, _hook_attribute, _hook_class


def known_hooks_key_to_object(key):
    """Convert a known_hooks key to a object.

    :param key: A tuple (module_name, member_name) as found in the keys of
        the known_hooks registry.
    :return: The object this specifies.
    """
    return pyutils.get_named_object(*key)


class Hooks(dict):
    """A dictionary mapping hook name to a list of callables.

    e.g. ['FOO'] Is the list of items to be called when the
    FOO hook is triggered.
    """

    def __init__(self, module=None, member_name=None):
        """Create a new hooks dictionary.

        :param module: The module from which this hooks dictionary should be loaded
            (used for lazy hooks)
        :param member_name: Name under which this hooks dictionary should be loaded.
            (used for lazy hooks)
        """
        dict.__init__(self)
        self._callable_names = {}
        self._lazy_callable_names = {}
        self._module = module
        self._member_name = member_name

    def add_hook(self, name, doc, introduced, deprecated=None):
        """Add a hook point to this dictionary.

        :param name: The name of the hook, for clients to use when registering.
        :param doc: The docs for the hook.
        :param introduced: When the hook was introduced (e.g. (0, 15)).
        :param deprecated: When the hook was deprecated, None for
            not-deprecated.
        """
        if name in self:
            raise errors.DuplicateKey(name)
        if self._module:
            callbacks = _lazy_hooks.setdefault(
                (self._module, self._member_name, name), [])
        else:
            callbacks = None
        hookpoint = HookPoint(name=name, doc=doc, introduced=introduced,
                              deprecated=deprecated, callbacks=callbacks)
        self[name] = hookpoint

    def docs(self):
        """Generate the documentation for this Hooks instance.

        This introspects all the individual hooks and returns their docs as well.
        """
        hook_names = sorted(self.keys())
        hook_docs = []
        name = self.__class__.__name__
        hook_docs.append(name)
        hook_docs.append("-" * len(name))
        hook_docs.append("")
        for hook_name in hook_names:
            hook = self[hook_name]
            hook_docs.append(hook.docs())
        return "\n".join(hook_docs)

    def get_hook_name(self, a_callable):
        """Get the name for a_callable for UI display.

        If no name has been registered, the string 'No hook name' is returned.
        We use a fixed string rather than repr or the callables module because
        the code names are rarely meaningful for end users and this is not
        intended for debugging.
        """
        name = self._callable_names.get(a_callable, None)
        if name is None and a_callable is not None:
            name = self._lazy_callable_names.get((a_callable.__module__,
                                                  a_callable.__name__),
                                                 None)
        if name is None:
            return 'No hook name'
        return name

    def install_named_hook_lazy(self, hook_name, callable_module,
                                callable_member, name):
        """Install a_callable in to the hook hook_name lazily, and label it.

        :param hook_name: A hook name. See the __init__ method for the complete
            list of hooks.
        :param callable_module: Name of the module in which the callable is
            present.
        :param callable_member: Member name of the callable.
        :param name: A name to associate the callable with, to show users what
            is running.
        """
        try:
            hook = self[hook_name]
        except KeyError:
            raise UnknownHook(self.__class__.__name__, hook_name)
        try:
            hook_lazy = getattr(hook, "hook_lazy")
        except AttributeError:
            raise errors.UnsupportedOperation(self.install_named_hook_lazy,
                                              self)
        else:
            hook_lazy(callable_module, callable_member, name)
        if name is not None:
            self.name_hook_lazy(callable_module, callable_member, name)

    def install_named_hook(self, hook_name, a_callable, name):
        """Install a_callable in to the hook hook_name, and label it name.

        :param hook_name: A hook name. See the __init__ method for the complete
            list of hooks.
        :param a_callable: The callable to be invoked when the hook triggers.
            The exact signature will depend on the hook - see the __init__
            method for details on each hook.
        :param name: A name to associate a_callable with, to show users what is
            running.
        """
        try:
            hook = self[hook_name]
        except KeyError:
            raise UnknownHook(self.__class__.__name__, hook_name)
        try:
            # list hooks, old-style, not yet deprecated but less useful.
            hook.append(a_callable)
        except AttributeError:
            hook.hook(a_callable, name)
        if name is not None:
            self.name_hook(a_callable, name)

    def uninstall_named_hook(self, hook_name, label):
        """Uninstall named hooks.

        :param hook_name: Hook point name
        :param label: Label of the callable to uninstall
        """
        try:
            hook = self[hook_name]
        except KeyError:
            raise UnknownHook(self.__class__.__name__, hook_name)
        try:
            uninstall = getattr(hook, "uninstall")
        except AttributeError:
            raise errors.UnsupportedOperation(self.uninstall_named_hook, self)
        else:
            uninstall(label)

    def name_hook(self, a_callable, name):
        """Associate name with a_callable to show users what is running."""
        self._callable_names[a_callable] = name

    def name_hook_lazy(self, callable_module, callable_member, callable_name):
        self._lazy_callable_names[(callable_module, callable_member)] = \
            callable_name


class HookPoint:
    """A single hook that clients can register to be called back when it fires.

    Attributes:
      name: The name of the hook.
      doc: The docs for using the hook.
      introduced: A version tuple specifying what version the hook was
                      introduced in. None indicates an unknown version.
      deprecated: A version tuple specifying what version the hook was
                      deprecated or superseded in. None indicates that the hook
                      is not superseded or deprecated. If the hook is
                      superseded then the doc should describe the recommended
                      replacement hook to register for.
    """

    def __init__(self, name, doc, introduced, deprecated=None, callbacks=None):
        """Create a HookPoint.

        :param name: The name of the hook, for clients to use when registering.
        :param doc: The docs for the hook.
        :param introduced: When the hook was introduced (e.g. (0, 15)).
        :param deprecated: When the hook was deprecated, None for
            not-deprecated.
        """
        self.name = name
        self.__doc__ = doc
        self.introduced = introduced
        self.deprecated = deprecated
        if callbacks is None:
            self._callbacks = []
        else:
            self._callbacks = callbacks

    def docs(self):
        """Generate the documentation for this HookPoint.

        :return: A string terminated in \n.
        """
        import textwrap
        strings = []
        strings.append(self.name)
        strings.append('~' * len(self.name))
        strings.append('')
        if self.introduced:
            introduced_string = _format_version_tuple(self.introduced)
        else:
            introduced_string = 'unknown'
        strings.append(gettext('Introduced in: %s') % introduced_string)
        if self.deprecated:
            deprecated_string = _format_version_tuple(self.deprecated)
            strings.append(gettext('Deprecated in: %s') % deprecated_string)
        strings.append('')
        strings.extend(textwrap.wrap(self.__doc__,
                                     break_long_words=False))
        strings.append('')
        return '\n'.join(strings)

    def __eq__(self, other):
        return (isinstance(other, type(self)) and other.__dict__ == self.__dict__)

    def hook_lazy(self, callback_module, callback_member, callback_label):
        """Lazily register a callback to be called when this HookPoint fires.

        :param callback_module: Module of the callable to use when this
            HookPoint fires.
        :param callback_member: Member name of the callback.
        :param callback_label: A label to show in the UI while this callback is
            processing.
        """
        obj_getter = registry._LazyObjectGetter(callback_module,
                                                callback_member)
        self._callbacks.append((obj_getter, callback_label))

    def hook(self, callback, callback_label):
        """Register a callback to be called when this HookPoint fires.

        :param callback: The callable to use when this HookPoint fires.
        :param callback_label: A label to show in the UI while this callback is
            processing.
        """
        obj_getter = registry._ObjectGetter(callback)
        self._callbacks.append((obj_getter, callback_label))

    def uninstall(self, label):
        """Uninstall the callback with the specified label.

        :param label: Label of the entry to uninstall
        """
        entries_to_remove = []
        for entry in self._callbacks:
            (entry_callback, entry_label) = entry
            if entry_label == label:
                entries_to_remove.append(entry)
        if entries_to_remove == []:
            raise KeyError("No entry with label %r" % label)
        for entry in entries_to_remove:
            self._callbacks.remove(entry)

    def __iter__(self):
        return (callback.get_obj() for callback, name in self._callbacks)

    def __len__(self):
        return len(self._callbacks)

    def __repr__(self):
        strings = []
        strings.append("<%s(" % type(self).__name__)
        strings.append(self.name)
        strings.append("), callbacks=[")
        callbacks = self._callbacks
        for (callback, callback_name) in callbacks:
            strings.append(repr(callback.get_obj()))
            strings.append("(")
            strings.append(callback_name)
            strings.append("),")
        if len(callbacks) == 1:
            strings[-1] = ")"
        strings.append("]>")
        return ''.join(strings)


_help_prefix = \
    """
Hooks
=====

Introduction
------------

A hook of type *xxx* of class *yyy* needs to be registered using::

  yyy.hooks.install_named_hook("xxx", ...)

See :doc:`Using hooks<../user-guide/hooks>` in the User Guide for examples.

The class that contains each hook is given before the hooks it supplies. For
instance, BranchHooks as the class is the hooks class for
`breezy.branch.Branch.hooks`.

Each description also indicates whether the hook runs on the client (the
machine where bzr was invoked) or the server (the machine addressed by
the branch URL).  These may be, but are not necessarily, the same machine.

Plugins (including hooks) are run on the server if all of these is true:

  * The connection is via a smart server (accessed with a URL starting with
    "bzr://", "bzr+ssh://" or "bzr+http://", or accessed via a "http://"
    URL when a smart server is available via HTTP).

  * The hook is either server specific or part of general infrastructure rather
    than client specific code (such as commit).

"""


def hooks_help_text(topic):
    segments = [_help_prefix]
    for hook_key in sorted(known_hooks.keys()):
        hooks = known_hooks_key_to_object(hook_key)
        segments.append(hooks.docs())
    return '\n'.join(segments)


# Lazily registered hooks. Maps (module, name, hook_name) tuples
# to lists of tuples with objectgetters and names
_lazy_hooks: Dict[Tuple[str, str, str], List[Tuple[registry._ObjectGetter, str]]] = {}


def install_lazy_named_hook(hookpoints_module, hookpoints_name, hook_name,
                            a_callable, name):
    """Install a callable in to a hook lazily, and label it name.

    :param hookpoints_module: Module name of the hook points.
    :param hookpoints_name: Name of the hook points.
    :param hook_name: A hook name.
    :param callable: a callable to call for the hook.
    :param name: A name to associate a_callable with, to show users what is
        running.
    """
    key = (hookpoints_module, hookpoints_name, hook_name)
    obj_getter = registry._ObjectGetter(a_callable)
    _lazy_hooks.setdefault(key, []).append((obj_getter, name))
